#!/usr/bin/env python
# -*- coding: utf-8 -*-
from dataclasses import dataclass
import numpy as np

try:
    import vedo.vtkclasses as vtk
except ImportError:
    import vtkmodules.all as vtk

__docformat__ = "google"

__doc__ = """Submodule to customize interaction modes."""


class MousePan(vtk.vtkInteractorStyleUser):
    """
    Interaction mode to pan the scene by dragging the mouse.

    Controls:
    - Left mouse button will pan the scene.
    - Mouse middle button up/down is elevation, and left and right is azimuth.
    - Right mouse button is rotate (left/right movement) and zoom in/out
      (up/down movement)
    - Mouse scroll wheel is zoom in/out
    """

    def __init__(self):

        super().__init__()

        self.left = False
        self.middle = False
        self.right = False

        self.interactor = None
        self.renderer = None
        self.camera = None

        self.oldpickD = []
        self.newpickD = []
        self.oldpickW = np.array([0, 0, 0, 0], dtype=float)
        self.newpickW = np.array([0, 0, 0, 0], dtype=float)
        self.fpD = np.array([0, 0, 0], dtype=float)
        self.fpW = np.array([0, 0, 0], dtype=float)
        self.posW = np.array([0, 0, 0], dtype=float)
        self.motionD = np.array([0, 0], dtype=float)
        self.motionW = np.array([0, 0, 0], dtype=float)

        self.AddObserver("LeftButtonPressEvent", self._left_down)
        self.AddObserver("LeftButtonReleaseEvent", self._left_up)
        self.AddObserver("MiddleButtonPressEvent", self._middle_down)
        self.AddObserver("MiddleButtonReleaseEvent", self._middle_up)
        self.AddObserver("RightButtonPressEvent", self._right_down)
        self.AddObserver("RightButtonReleaseEvent", self._right_up)
        self.AddObserver("MouseWheelForwardEvent", self._wheel_forward)
        self.AddObserver("MouseWheelBackwardEvent", self._wheel_backward)
        self.AddObserver("MouseMoveEvent", self._mouse_move)

    def _get_motion(self):
        self.oldpickD = np.array(self.interactor.GetLastEventPosition())
        self.newpickD = np.array(self.interactor.GetEventPosition())
        self.motionD = (self.newpickD - self.oldpickD) / 4
        self.camera = self.renderer.GetActiveCamera()
        self.fpW = self.camera.GetFocalPoint()
        self.posW = self.camera.GetPosition()
        self.ComputeWorldToDisplay(self.renderer, self.fpW[0], self.fpW[1], self.fpW[2], self.fpD)
        focaldepth = self.fpD[2]
        self.ComputeDisplayToWorld(
            self.renderer, self.oldpickD[0], self.oldpickD[1], focaldepth, self.oldpickW
        )
        self.ComputeDisplayToWorld(
            self.renderer, self.newpickD[0], self.newpickD[1], focaldepth, self.newpickW
        )
        self.motionW[:3] = self.oldpickW[:3] - self.newpickW[:3]

    def _mouse_left_move(self):
        self._get_motion()
        self.camera.SetFocalPoint(self.fpW[:3] + self.motionW[:3])
        self.camera.SetPosition(self.posW[:3] + self.motionW[:3])
        self.interactor.Render()

    def _mouse_middle_move(self):
        self._get_motion()
        if abs(self.motionD[0]) > abs(self.motionD[1]):
            self.camera.Azimuth(-2 * self.motionD[0])
        else:
            self.camera.Elevation(-self.motionD[1])
        self.interactor.Render()

    def _mouse_right_move(self):
        self._get_motion()
        if abs(self.motionD[0]) > abs(self.motionD[1]):
            self.camera.Azimuth(-2.0 * self.motionD[0])
        else:
            self.camera.Zoom(1 + self.motionD[1] / 100)
        self.interactor.Render()

    def _mouse_wheel_forward(self):
        self.camera = self.renderer.GetActiveCamera()
        self.camera.Zoom(1.1)
        self.interactor.Render()

    def _mouse_wheel_backward(self):
        self.camera = self.renderer.GetActiveCamera()
        self.camera.Zoom(0.9)
        self.interactor.Render()

    def _left_down(self, w, e):
        self.left = True

    def _left_up(self, w, e):
        self.left = False

    def _middle_down(self, w, e):
        self.middle = True

    def _middle_up(self, w, e):
        self.middle = False

    def _right_down(self, w, e):
        self.right = True

    def _right_up(self, w, e):
        self.right = False

    def _wheel_forward(self, w, e):
        self._mouse_wheel_forward()

    def _wheel_backward(self, w, e):
        self._mouse_wheel_backward()

    def _mouse_move(self, w, e):
        if self.left:
            self._mouse_left_move()
        if self.middle:
            self._mouse_middle_move()
        if self.right:
            self._mouse_right_move()


###################################################################################
@dataclass
class _BlenderStyleDragInfo:
    """Data structure containing the data required to execute dragging a node"""

    # Scene related
    dragged_node = None  # Node

    # VTK related
    actors_dragging: list
    dragged_actors_original_positions: list  # original VTK positions
    start_position_3d = np.array((0, 0, 0))  # start position of the cursor

    delta = np.array((0, 0, 0))

    def __init__(self):
        self.actors_dragging = []
        self.dragged_actors_original_positions = []


###############################################
class BlenderStyle(vtk.vtkInteractorStyleUser):
    """
    Create an interaction style using the Blender default key-bindings.

    Camera action code is largely a translation of
    [this](https://github.com/Kitware/VTK/blob/master/Interaction/Style/vtkInteractorStyleTrackballCamera.cxx)
    Rubber band code
    [here](https://gitlab.kitware.com/updega2/vtk/-/blob/d324b2e898b0da080edee76159c2f92e6f71abe2/Rendering/vtkInteractorStyleRubberBandZoom.cxx)

    Interaction:

    Left button: Sections
    ----------------------
    Left button: select

    Left button drag: rubber band select or line select, depends on the dragged distance

    Middle button: Navigation
    --------------------------
    Middle button: rotate

    Middle button + shift : pan

    Middle button + ctrl  : zoom

    Middle button + alt : center view on picked point

    OR

    Middle button + alt   : zoom rubber band

    Mouse wheel : zoom

    Right button : context
    -----------------------
    Right key click: reserved for context-menu


    Keys
    ----

    2 or 3 : toggle perspective view

    a      : zoom all

    x,y,z  : view direction (toggles positive and negative)

    left/right arrows: rotate 45 deg clockwise/ccw about z-axis, snaps to nearest 45 deg
    b      : box zoom

    m      : mouse middle lock (toggles)

    space  : same as middle mouse button

    g      : grab (move actors)

    enter  : accept drag

    esc    : cancel drag, call callbackEscape


    LAPTOP MODE
    -----------
    Use space or `m` as replacement for middle button
    (`m` is sticky, space is not)

    callbacks / overriding keys:

    if `callbackAnyKey` is assigned then this function is called on every key press.
    If this function returns True then further processing of events is stopped.


    Moving actors
    --------------
    Actors can be moved interactively by the user.
    To support custom groups of actors to be moved as a whole the following system
    is implemented:

    When 'g' is pressed (grab) then a `_BlenderStyleDragInfo` dataclass object is assigned
    to style to `style.draginfo`.

    `_BlenderStyleDragInfo` includes a list of all the actors that are being dragged.
    By default this is the selection, but this may be altered.
    Drag is accepted using enter, click, or g. Drag is cancelled by esc

    Events
    ------
    `callbackStartDrag` is called when initializing the drag.
    This is when to assign actors and other data to draginfo.

    `callbackEndDrag` is called when the drag is accepted.

    Responding to other events
    --------------------------
    `callbackCameraDirectionChanged` : executed when camera has rotated but before re-rendering

    .. note::
        This class is based on R. de Bruin's
        [DAVE](https://github.com/RubendeBruin/DAVE/blob/master/src/DAVE/visual_helpers/vtkBlenderLikeInteractionStyle.py)
        implementation as discussed in this
        [issue](https://github.com/marcomusy/vedo/discussions/788).

    Example:
        ```python
        from vedo import *
        settings.enable_default_keyboard_callbacks = False
        settings.enable_default_mouse_callbacks = False
        mesh = Mesh(dataurl+"cow.vtk")
        mode = interactor_modes.BlenderStyle()
        plt = Plotter().user_mode(mode)
        plt.show(mesh, axes=1)
        ```
    """

    def __init__(self):

        super().__init__()

        self.interactor = None
        self.renderer = None

        # callbackSelect is called whenever one or mode props are selected.
        # callback will be called with a list of props of which the first entry
        # is prop closest to the camera.
        self.callbackSelect = None
        self.callbackStartDrag = None
        self.callbackEndDrag = None
        self.callbackEscapeKey = None
        self.callbackDeleteKey = None
        self.callbackFocusKey = None
        self.callbackAnyKey = None
        self.callbackMeasure = None  # callback with argument float (meters)
        self.callbackCameraDirectionChanged = None

        # active drag
        # assigned to a _BlenderStyleDragInfo object when dragging is active
        self.draginfo: _BlenderStyleDragInfo or None = None

        # picking
        self.picked_props = []  # will be filled by latest pick

        # settings
        self.mouse_motion_factor = 20
        self.mouse_wheel_motion_factor = 0.1
        self.zoom_motion_factor = 0.25

        # internals
        self.start_x = 0  # start of a drag
        self.start_y = 0
        self.end_x = 0
        self.end_y = 0

        self.middle_mouse_lock = False
        self.middle_mouse_lock_actor = None  # will be created when required

        # Special Modes
        self._is_box_zooming = False

        # holds an image of the renderer output at the start of a drawing event
        self._pixel_array = vtk.vtkUnsignedCharArray()

        self._upside_down = False

        self._left_button_down = False
        self._middle_button_down = False

        self.AddObserver("RightButtonPressEvent", self.RightButtonPress)
        self.AddObserver("RightButtonReleaseEvent", self.RightButtonRelease)
        self.AddObserver("MiddleButtonPressEvent", self.MiddleButtonPress)
        self.AddObserver("MiddleButtonReleaseEvent", self.MiddleButtonRelease)
        self.AddObserver("MouseWheelForwardEvent", self.MouseWheelForward)
        self.AddObserver("MouseWheelBackwardEvent", self.MouseWheelBackward)
        self.AddObserver("LeftButtonPressEvent", self.LeftButtonPress)
        self.AddObserver("LeftButtonReleaseEvent", self.LeftButtonRelease)
        self.AddObserver("MouseMoveEvent", self.MouseMove)
        self.AddObserver("WindowResizeEvent", self.WindowResized)
        # ^does not seem to fire!
        self.AddObserver("KeyPressEvent", self.KeyPress)
        self.AddObserver("KeyReleaseEvent", self.KeyRelease)

    def RightButtonPress(self, obj, event):
        pass

    def RightButtonRelease(self, obj, event):
        pass

    def MiddleButtonPress(self, obj, event):
        self._middle_button_down = True

    def MiddleButtonRelease(self, obj, event):
        self._middle_button_down = False

        # perform middle button focus event if ALT is down
        if self.GetInteractor().GetAltKey():
            # print("Middle button released while ALT is down")

            # try to pick an object at the current mouse position
            rwi = self.GetInteractor()
            self.start_x, self.start_y = rwi.GetEventPosition()
            props = self.PerformPickingOnSelection()

            if props:
                self.FocusOn(props[0])

    def MouseWheelBackward(self, obj, event):
        self.MoveMouseWheel(-1)

    def MouseWheelForward(self, obj, event):
        self.MoveMouseWheel(1)

    def MouseMove(self, obj, event):

        interactor = self.GetInteractor()

        # Find the renderer that is active below the current mouse position
        x, y = interactor.GetEventPosition()
        self.FindPokedRenderer(x, y)
        # sets the current renderer
        # [this->SetCurrentRenderer(this->Interactor->FindPokedRenderer(x, y));]

        Shift = interactor.GetShiftKey()
        Ctrl = interactor.GetControlKey()
        Alt = interactor.GetAltKey()

        MiddleButton = self._middle_button_down or self.middle_mouse_lock

        # start with the special modes
        if self._is_box_zooming:
            self.DrawDraggedSelection()
        elif MiddleButton and not Shift and not Ctrl and not Alt:
            self.Rotate()
        elif MiddleButton and Shift and not Ctrl and not Alt:
            self.Pan()
        elif MiddleButton and Ctrl and not Shift and not Alt:
            self.Zoom()  # Dolly
        elif self.draginfo is not None:
            self.ExecuteDrag()
        elif self._left_button_down and Ctrl and Shift:
            self.DrawMeasurement()
        elif self._left_button_down:
            self.DrawDraggedSelection()

        self.InvokeEvent("InteractionEvent", None)

    def MoveMouseWheel(self, direction):
        rwi = self.GetInteractor()

        # Find the renderer that is active below the current mouse position
        x, y = rwi.GetEventPosition()
        self.FindPokedRenderer(x, y)
        # sets the current renderer
        # [this->SetCurrentRenderer(this->Interactor->FindPokedRenderer(x, y));]

        # The movement

        CurrentRenderer = self.GetCurrentRenderer()

        #   // Calculate the focal depth since we'll be using it a lot
        camera = CurrentRenderer.GetActiveCamera()
        viewFocus = camera.GetFocalPoint()

        temp_out = [0, 0, 0]
        self.ComputeWorldToDisplay(
            CurrentRenderer, viewFocus[0], viewFocus[1], viewFocus[2], temp_out
        )
        focalDepth = temp_out[2]

        newPickPoint = [0, 0, 0, 0]
        x, y = rwi.GetEventPosition()
        self.ComputeDisplayToWorld(CurrentRenderer, x, y, focalDepth, newPickPoint)

        #   // Has to recalc old mouse point since the viewport has moved,
        #   // so can't move it outside the loop

        oldPickPoint = [0, 0, 0, 0]
        # xp, yp = rwi.GetLastEventPosition()

        # find the center of the window
        size = rwi.GetRenderWindow().GetSize()
        xp = size[0] / 2
        yp = size[1] / 2

        self.ComputeDisplayToWorld(CurrentRenderer, xp, yp, focalDepth, oldPickPoint)
        #
        #   // Camera motion is reversed
        #
        move_factor = -1 * self.zoom_motion_factor * direction

        motionVector = (
            move_factor * (oldPickPoint[0] - newPickPoint[0]),
            move_factor * (oldPickPoint[1] - newPickPoint[1]),
            move_factor * (oldPickPoint[2] - newPickPoint[2]),
        )

        viewFocus = camera.GetFocalPoint()  # do we need to do this again? Already did this
        viewPoint = camera.GetPosition()

        camera.SetFocalPoint(
            motionVector[0] + viewFocus[0],
            motionVector[1] + viewFocus[1],
            motionVector[2] + viewFocus[2],
        )
        camera.SetPosition(
            motionVector[0] + viewPoint[0],
            motionVector[1] + viewPoint[1],
            motionVector[2] + viewPoint[2],
        )

        # the Zooming
        factor = self.mouse_motion_factor * self.mouse_wheel_motion_factor
        self.ZoomByStep(direction * factor)

    def ZoomByStep(self, step):
        CurrentRenderer = self.GetCurrentRenderer()

        if CurrentRenderer:
            self.StartDolly()
            self.Dolly(pow(1.1, step))
            self.EndDolly()

    def LeftButtonPress(self, obj, event):

        if self._is_box_zooming:
            return
        if self.draginfo:
            return

        self._left_button_down = True

        interactor = self.GetInteractor()
        Shift = interactor.GetShiftKey()
        Ctrl = interactor.GetControlKey()

        if Shift and Ctrl:
            if not self.GetCurrentRenderer().GetActiveCamera().GetParallelProjection():
                self.ToggleParallelProjection()

        rwi = self.GetInteractor()
        self.start_x, self.start_y = rwi.GetEventPosition()
        self.end_x = self.start_x
        self.end_y = self.start_y

        self.InitializeScreenDrawing()

    def LeftButtonRelease(self, obj, event):

        if self._is_box_zooming:
            self._is_box_zooming = False
            self.ZoomBox(self.start_x, self.start_y, self.end_x, self.end_y)
            return

        if self.draginfo:
            self.FinishDrag()
            return

        self._left_button_down = False

        interactor = self.GetInteractor()

        Shift = interactor.GetShiftKey()
        Ctrl = interactor.GetControlKey()
        Alt = interactor.GetAltKey()

        if Ctrl and Shift:
            pass  # we were drawing the measurement

        else:
            if self.callbackSelect:
                props = self.PerformPickingOnSelection()

                if props:  # only call back if anything was selected
                    self.picked_props = tuple(props)
                    self.callbackSelect(props)

        # remove the selection rubber band / line
        self.DoRender()

    def KeyPress(self, obj, event):

        key = obj.GetKeySym()
        KEY = key.upper()

        # logging.info(f"Key Press: {key}")
        if self.callbackAnyKey:
            if self.callbackAnyKey(key):
                return

        if KEY == "M":
            self.middle_mouse_lock = not self.middle_mouse_lock
            self.UpdateMiddleMouseButtonLockActor()
        elif KEY == "G":
            if self.draginfo is not None:
                self.FinishDrag()
            else:
                if self.callbackStartDrag:
                    self.callbackStartDrag()
                else:
                    self.StartDrag()
                    # internally calls end-drag if drag is already active
        elif KEY == "ESCAPE":
            if self.callbackEscapeKey:
                self.callbackEscapeKey()
            if self.draginfo is not None:
                self.CancelDrag()
        elif KEY == "DELETE":
            if self.callbackDeleteKey:
                self.callbackDeleteKey()
        elif KEY == "RETURN":
            if self.draginfo:
                self.FinishDrag()
        elif KEY == "SPACE":
            self.middle_mouse_lock = True
            # self.UpdateMiddleMouseButtonLockActor()
            # self.GrabFocus("MouseMoveEvent", self)
            # # TODO: grab and release focus; possible from python?
        elif KEY == "B":
            self._is_box_zooming = True
            rwi = self.GetInteractor()
            self.start_x, self.start_y = rwi.GetEventPosition()
            self.end_x = self.start_x
            self.end_y = self.start_y
            self.InitializeScreenDrawing()
        elif KEY in ('2', '3'):
            self.ToggleParallelProjection()

        elif KEY == "A":
            self.ZoomFit()
        elif KEY == "X":
            self.SetViewX()
        elif KEY == "Y":
            self.SetViewY()
        elif KEY == "Z":
            self.SetViewZ()
        elif KEY == "LEFT":
            self.RotateDiscreteStep(1)
        elif KEY == "RIGHT":
            self.RotateDiscreteStep(-1)
        elif KEY == "UP":
            self.RotateTurtableBy(0, 10)
        elif KEY == "DOWN":
            self.RotateTurtableBy(0, -10)
        elif KEY == "PLUS":
            self.ZoomByStep(2)
        elif KEY == "MINUS":
            self.ZoomByStep(-2)
        elif KEY == "F":
            if self.callbackFocusKey:
                self.callbackFocusKey()

        self.InvokeEvent("InteractionEvent", None)

    def KeyRelease(self, obj, event):

        key = obj.GetKeySym()
        KEY = key.upper()

        # print(f"Key release: {key}")

        if KEY == "SPACE":
            if self.middle_mouse_lock:
                self.middle_mouse_lock = False
                self.UpdateMiddleMouseButtonLockActor()

    def WindowResized(self):
        # print("window resized")
        self.InitializeScreenDrawing()

    def RotateDiscreteStep(self, movement_direction, step=22.5):
        """Rotates CW or CCW to the nearest 45 deg angle
        - includes some fuzzyness to determine about which axis"""

        CurrentRenderer = self.GetCurrentRenderer()
        camera = CurrentRenderer.GetActiveCamera()

        step = np.deg2rad(step)

        direction = -np.array(camera.GetViewPlaneNormal())  # current camera direction

        if abs(direction[2]) < 0.7:
            # horizontal view, rotate camera position about Z-axis
            angle = np.arctan2(direction[1], direction[0])

            # find the nearest angle that is an integer number of steps
            if movement_direction > 0:
                angle = step * np.floor((angle + 0.1 * step) / step) + step
            else:
                angle = -step * np.floor(-(angle - 0.1 * step) / step) - step

            dist = np.linalg.norm(direction[:2])

            direction[0] = np.cos(angle) * dist
            direction[1] = np.sin(angle) * dist

            self.SetCameraDirection(direction)

        else:  # Top or bottom like view - rotate camera "up" direction

            up = np.array(camera.GetViewUp())

            angle = np.arctan2(up[1], up[0])

            # find the nearest angle that is an integer number of steps
            if movement_direction > 0:
                angle = step * np.floor((angle + 0.1 * step) / step) + step
            else:
                angle = -step * np.floor(-(angle - 0.1 * step) / step) - step

            dist = np.linalg.norm(up[:2])

            up[0] = np.cos(angle) * dist
            up[1] = np.sin(angle) * dist

            camera.SetViewUp(up)
            camera.OrthogonalizeViewUp()

            self.DoRender()

    def ToggleParallelProjection(self):
        renderer = self.GetCurrentRenderer()
        camera = renderer.GetActiveCamera()
        camera.SetParallelProjection(not bool(camera.GetParallelProjection()))
        self.DoRender()

    def SetViewX(self):
        self.SetCameraPlaneDirection((1, 0, 0))

    def SetViewY(self):
        self.SetCameraPlaneDirection((0, 1, 0))

    def SetViewZ(self):
        self.SetCameraPlaneDirection((0, 0, 1))

    def ZoomFit(self):
        self.GetCurrentRenderer().ResetCamera()
        self.DoRender()

    def SetCameraPlaneDirection(self, direction):
        """Sets the camera to display a plane of which direction is the normal
        - includes logic to reverse the direction if benificial"""

        CurrentRenderer = self.GetCurrentRenderer()
        camera = CurrentRenderer.GetActiveCamera()

        direction = np.array(direction)

        normal = camera.GetViewPlaneNormal()
        # can not set the normal, need to change the position to do that

        current_alignment = np.dot(normal, -direction)
        # print(f"Current alignment = {current_alignment}")

        if abs(current_alignment) > 0.9999:
            # print("toggling")
            direction = -np.array(normal)
        elif current_alignment > 0:  # find the nearest plane
            # print("reversing to find nearest")
            direction = -direction

        self.SetCameraDirection(-direction)

    def SetCameraDirection(self, direction):
        """Sets the camera to this direction, sets view up if horizontal enough"""
        direction = np.array(direction)

        CurrentRenderer = self.GetCurrentRenderer()
        camera = CurrentRenderer.GetActiveCamera()
        rwi = self.GetInteractor()

        pos = np.array(camera.GetPosition())
        focal = np.array(camera.GetFocalPoint())
        dist = np.linalg.norm(pos - focal)

        pos = focal - dist * direction
        camera.SetPosition(pos)

        if abs(direction[2]) < 0.9:
            camera.SetViewUp(0, 0, 1)
        elif direction[2] > 0.9:
            camera.SetViewUp(0, -1, 0)
        else:
            camera.SetViewUp(0, 1, 0)

        camera.OrthogonalizeViewUp()

        if self.GetAutoAdjustCameraClippingRange():
            CurrentRenderer.ResetCameraClippingRange()

        if rwi.GetLightFollowCamera():
            CurrentRenderer.UpdateLightsGeometryToFollowCamera()

        if self.callbackCameraDirectionChanged:
            self.callbackCameraDirectionChanged()

        self.DoRender()

    def PerformPickingOnSelection(self):
        """Preforms prop3d picking on the current dragged selection

        If the distance between the start and endpoints is less than the threshold
        then a SINGLE prop3d is picked along the line

        the selection area is drawn by the rubber band and is defined by
        self.start_x, self.start_y, self.end_x, self.end_y
        """
        renderer = self.GetCurrentRenderer()

        assemblyPath = renderer.PickProp(self.start_x, self.start_y, self.end_x, self.end_y)

        # re-pick in larger area if nothing is returned
        if not assemblyPath:
            self.start_x -= 2
            self.end_x += 2
            self.start_y -= 2
            self.end_y += 2
            assemblyPath = renderer.PickProp(self.start_x, self.start_y, self.end_x, self.end_y)

        # The nearest prop (by Z-value)
        if assemblyPath:
            assert (
                assemblyPath.GetNumberOfItems() == 1
            ), "Wrong assumption on number of returned nodes when picking"
            nearest_prop = assemblyPath.GetItemAsObject(0).GetViewProp()

            # all props
            collection = renderer.GetPickResultProps()
            props = [collection.GetItemAsObject(i) for i in range(collection.GetNumberOfItems())]

            props.remove(nearest_prop)
            props.insert(0, nearest_prop)

            return props

        else:
            return []

    # ----------- actor dragging ------------

    def StartDrag(self):
        if self.callbackStartDrag:
            # print("Calling callbackStartDrag")
            self.callbackStartDrag()
            return
        else:  # grab the current selection
            if self.picked_props:
                self.StartDragOnProps(self.picked_props)
            else:
                pass
                # print('Can not start drag, nothing selected and callbackStartDrag not assigned')

    def FinishDrag(self):
        # print('Finished drag')
        if self.callbackEndDrag:
            # reset actor positions as actors positions will be controlled by called functions
            for pos0, actor in zip(
                self.draginfo.dragged_actors_original_positions, self.draginfo.actors_dragging
            ):
                actor.SetPosition(pos0)
            self.callbackEndDrag(self.draginfo)

        self.draginfo = None

    def StartDragOnProps(self, props):
        """Starts drag on the provided props (actors) by filling self.draginfo"""
        if self.draginfo is not None:
            self.FinishDrag()
            return

        # print('Starting drag')

        # create and fill drag-info
        draginfo = _BlenderStyleDragInfo()

        #
        # draginfo.dragged_node = node
        #
        # # find all actors and outlines corresponding to this node
        # actors = [*self.actor_from_node(node).actors.values()]
        # outlines = [ol.outline_actor for ol in self.node_outlines if ol.parent_vp_actor in actors]

        draginfo.actors_dragging = props  # [*actors, *outlines]

        for a in draginfo.actors_dragging:
            draginfo.dragged_actors_original_positions.append(a.GetPosition())  # numpy ndarray

        # Get the start position of the drag in 3d

        rwi = self.GetInteractor()
        CurrentRenderer = self.GetCurrentRenderer()
        camera = CurrentRenderer.GetActiveCamera()
        viewFocus = camera.GetFocalPoint()

        temp_out = [0, 0, 0]
        self.ComputeWorldToDisplay(
            CurrentRenderer, viewFocus[0], viewFocus[1], viewFocus[2], temp_out
        )
        focalDepth = temp_out[2]

        newPickPoint = [0, 0, 0, 0]
        x, y = rwi.GetEventPosition()
        self.ComputeDisplayToWorld(CurrentRenderer, x, y, focalDepth, newPickPoint)

        mouse_pos_3d = np.array(newPickPoint[:3])

        draginfo.start_position_3d = mouse_pos_3d

        self.draginfo = draginfo

    def ExecuteDrag(self):

        rwi = self.GetInteractor()
        CurrentRenderer = self.GetCurrentRenderer()

        camera = CurrentRenderer.GetActiveCamera()
        viewFocus = camera.GetFocalPoint()

        # Get the picked point in 3d

        temp_out = [0, 0, 0]
        self.ComputeWorldToDisplay(
            CurrentRenderer, viewFocus[0], viewFocus[1], viewFocus[2], temp_out
        )
        focalDepth = temp_out[2]

        newPickPoint = [0, 0, 0, 0]
        x, y = rwi.GetEventPosition()
        self.ComputeDisplayToWorld(CurrentRenderer, x, y, focalDepth, newPickPoint)

        mouse_pos_3d = np.array(newPickPoint[:3])

        # compute the delta and execute

        delta = np.array(mouse_pos_3d) - self.draginfo.start_position_3d
        # print(f'Delta = {delta}')
        view_normal = np.array(self.GetCurrentRenderer().GetActiveCamera().GetViewPlaneNormal())

        delta_inplane = delta - view_normal * np.dot(delta, view_normal)
        # print(f'delta_inplane = {delta_inplane}')

        for pos0, actor in zip(
            self.draginfo.dragged_actors_original_positions, self.draginfo.actors_dragging
        ):
            m = actor.GetUserMatrix()
            if m:
                print("UserMatrices/transforms not supported")
                # m.Invert() #inplace
                # rotated = m.MultiplyFloatPoint([*delta_inplane, 1])
                # actor.SetPosition(pos0 + np.array(rotated[:3]))
            actor.SetPosition(pos0 + delta_inplane)

        # print(f'Set position to {pos0 + delta_inplane}')

        self.draginfo.delta = delta_inplane  # store the current delta

        # self.GetInteractor().Render()
        self.DoRender()

    def CancelDrag(self):
        """Cancels the drag and restored the original positions of all dragged actors"""
        for pos0, actor in zip(
            self.draginfo.dragged_actors_original_positions, self.draginfo.actors_dragging
        ):
            actor.SetPosition(pos0)
        self.draginfo = None
        self.DoRender()

    # ----------- end dragging --------------

    def Zoom(self):
        rwi = self.GetInteractor()
        x, y = rwi.GetEventPosition()
        xp, yp = rwi.GetLastEventPosition()

        direction = y - yp
        self.MoveMouseWheel(direction / 10)

    def Pan(self):

        CurrentRenderer = self.GetCurrentRenderer()

        if CurrentRenderer:

            rwi = self.GetInteractor()

            #   // Calculate the focal depth since we'll be using it a lot
            camera = CurrentRenderer.GetActiveCamera()
            viewFocus = camera.GetFocalPoint()

            temp_out = [0, 0, 0]
            self.ComputeWorldToDisplay(
                CurrentRenderer, viewFocus[0], viewFocus[1], viewFocus[2], temp_out
            )
            focalDepth = temp_out[2]

            newPickPoint = [0, 0, 0, 0]
            x, y = rwi.GetEventPosition()
            self.ComputeDisplayToWorld(CurrentRenderer, x, y, focalDepth, newPickPoint)

            #   // Has to recalc old mouse point since the viewport has moved,
            #   // so can't move it outside the loop

            oldPickPoint = [0, 0, 0, 0]
            xp, yp = rwi.GetLastEventPosition()
            self.ComputeDisplayToWorld(CurrentRenderer, xp, yp, focalDepth, oldPickPoint)
            #
            #   // Camera motion is reversed
            #
            motionVector = (
                oldPickPoint[0] - newPickPoint[0],
                oldPickPoint[1] - newPickPoint[1],
                oldPickPoint[2] - newPickPoint[2],
            )

            viewFocus = camera.GetFocalPoint()  # do we need to do this again? Already did this
            viewPoint = camera.GetPosition()

            camera.SetFocalPoint(
                motionVector[0] + viewFocus[0],
                motionVector[1] + viewFocus[1],
                motionVector[2] + viewFocus[2],
            )
            camera.SetPosition(
                motionVector[0] + viewPoint[0],
                motionVector[1] + viewPoint[1],
                motionVector[2] + viewPoint[2],
            )

            if rwi.GetLightFollowCamera():
                CurrentRenderer.UpdateLightsGeometryToFollowCamera()

            self.DoRender()

    def Rotate(self):

        CurrentRenderer = self.GetCurrentRenderer()

        if CurrentRenderer:

            rwi = self.GetInteractor()
            dx = rwi.GetEventPosition()[0] - rwi.GetLastEventPosition()[0]
            dy = rwi.GetEventPosition()[1] - rwi.GetLastEventPosition()[1]

            size = CurrentRenderer.GetRenderWindow().GetSize()
            delta_elevation = -20.0 / size[1]
            delta_azimuth = -20.0 / size[0]

            rxf = dx * delta_azimuth * self.mouse_motion_factor
            ryf = dy * delta_elevation * self.mouse_motion_factor

            self.RotateTurtableBy(rxf, ryf)

    def RotateTurtableBy(self, rxf, ryf):

        CurrentRenderer = self.GetCurrentRenderer()
        rwi = self.GetInteractor()

        # rfx is rotation about the global Z vector (turn-table mode)
        # rfy is rotation about the side vector

        camera = CurrentRenderer.GetActiveCamera()
        campos = np.array(camera.GetPosition())
        focal = np.array(camera.GetFocalPoint())
        up = camera.GetViewUp()
        upside_down_factor = -1 if up[2] < 0 else 1

        # rotate about focal point

        P = campos - focal  # camera position

        # Rotate left/right about the global Z axis
        H = np.linalg.norm(P[:2])  # horizontal distance of camera to focal point
        elev = np.arctan2(P[2], H)  # elevation

        # if the camera is near the poles, then derive the azimuth from the up-vector
        sin_elev = np.sin(elev)
        if abs(sin_elev) < 0.8:
            azi = np.arctan2(P[1], P[0])  # azimuth from camera position
        else:
            if sin_elev < -0.8:
                azi = np.arctan2(upside_down_factor * up[1], upside_down_factor * up[0])
            else:
                azi = np.arctan2(-upside_down_factor * up[1], -upside_down_factor * up[0])

        D = np.linalg.norm(P)  # distance from focal point to camera

        # apply the change in azimuth and elevation
        azi_new = azi + rxf / 60

        elev_new = elev + upside_down_factor * ryf / 60

        # the changed elevation changes H (D stays the same)
        Hnew = D * np.cos(elev_new)

        # calculate new camera position relative to focal point
        Pnew = np.array((Hnew * np.cos(azi_new), Hnew * np.sin(azi_new), D * np.sin(elev_new)))

        # calculate the up-direction of the camera
        up_z = upside_down_factor * np.cos(elev_new)  # z follows directly from elevation
        up_h = upside_down_factor * np.sin(elev_new)  # horizontal component
        #
        # if upside_down:
        #     up_z = -up_z
        #     up_h = -up_h

        up = (-up_h * np.cos(azi_new), -up_h * np.sin(azi_new), up_z)

        new_pos = focal + Pnew

        camera.SetViewUp(up)
        camera.SetPosition(new_pos)

        camera.OrthogonalizeViewUp()

        # Update

        if self.GetAutoAdjustCameraClippingRange():
            CurrentRenderer.ResetCameraClippingRange()

        if rwi.GetLightFollowCamera():
            CurrentRenderer.UpdateLightsGeometryToFollowCamera()

        if self.callbackCameraDirectionChanged:
            self.callbackCameraDirectionChanged()

        self.DoRender()

    def ZoomBox(self, x1, y1, x2, y2):
        """Zooms to a box"""
        # int width, height;
        #   width = abs(this->EndPosition[0] - this->StartPosition[0]);
        #   height = abs(this->EndPosition[1] - this->StartPosition[1]);

        if x1 > x2:
            _ = x1
            x1 = x2
            x2 = _
        if y1 > y2:
            _ = y1
            y1 = y2
            y2 = _

        width = x2 - x1
        height = y2 - y1

        #   int *size = this->CurrentRenderer->GetSize();
        CurrentRenderer = self.GetCurrentRenderer()
        size = CurrentRenderer.GetSize()
        origin = CurrentRenderer.GetOrigin()
        camera = CurrentRenderer.GetActiveCamera()

        # Assuming we're drawing the band on the view-plane
        rbcenter = (x1 + width / 2, y1 + height / 2, 0)

        CurrentRenderer.SetDisplayPoint(rbcenter)
        CurrentRenderer.DisplayToView()
        CurrentRenderer.ViewToWorld()

        worldRBCenter = CurrentRenderer.GetWorldPoint()

        invw = 1.0 / worldRBCenter[3]
        worldRBCenter = [c * invw for c in worldRBCenter]
        winCenter = [origin[0] + 0.5 * size[0], origin[1] + 0.5 * size[1], 0]

        CurrentRenderer.SetDisplayPoint(winCenter)
        CurrentRenderer.DisplayToView()
        CurrentRenderer.ViewToWorld()

        worldWinCenter = CurrentRenderer.GetWorldPoint()
        invw = 1.0 / worldWinCenter[3]
        worldWinCenter = [c * invw for c in worldWinCenter]

        translation = [
            worldRBCenter[0] - worldWinCenter[0],
            worldRBCenter[1] - worldWinCenter[1],
            worldRBCenter[2] - worldWinCenter[2],
        ]

        pos = camera.GetPosition()
        fp = camera.GetFocalPoint()
        #
        pos = [pos[i] + translation[i] for i in range(3)]
        fp = [fp[i] + translation[i] for i in range(3)]

        #
        camera.SetPosition(pos)
        camera.SetFocalPoint(fp)

        if width > height:
            if width:
                camera.Zoom(size[0] / width)
        else:
            if height:
                camera.Zoom(size[1] / height)

        self.DoRender()

    def FocusOn(self, prop3D):
        """Move the camera to focus on this particular prop3D"""

        position = prop3D.GetPosition()

        # print(f"Focus on {position}")

        CurrentRenderer = self.GetCurrentRenderer()
        camera = CurrentRenderer.GetActiveCamera()

        fp = camera.GetFocalPoint()
        pos = camera.GetPosition()

        camera.SetFocalPoint(position)
        camera.SetPosition(
            position[0] - fp[0] + pos[0],
            position[1] - fp[1] + pos[1],
            position[2] - fp[2] + pos[2],
        )

        if self.GetAutoAdjustCameraClippingRange():
            CurrentRenderer.ResetCameraClippingRange()

        rwi = self.GetInteractor()
        if rwi.GetLightFollowCamera():
            CurrentRenderer.UpdateLightsGeometryToFollowCamera()

        self.DoRender()

    def Dolly(self, factor):
        CurrentRenderer = self.GetCurrentRenderer()

        if CurrentRenderer:
            camera = CurrentRenderer.GetActiveCamera()

            if camera.GetParallelProjection():
                camera.SetParallelScale(camera.GetParallelScale() / factor)
            else:
                camera.Dolly(factor)
                if self.GetAutoAdjustCameraClippingRange():
                    CurrentRenderer.ResetCameraClippingRange()

            # if not do_not_update:
            #     rwi = self.GetInteractor()
            #     if rwi.GetLightFollowCamera():
            #         CurrentRenderer.UpdateLightsGeometryToFollowCamera()
            #     # rwi.Render()
            #     self.DoRender()

    def DrawMeasurement(self):
        rwi = self.GetInteractor()
        self.end_x, self.end_y = rwi.GetEventPosition()
        self.DrawLine(self.start_x, self.end_x, self.start_y, self.end_y)

    def DrawDraggedSelection(self):
        rwi = self.GetInteractor()
        self.end_x, self.end_y = rwi.GetEventPosition()
        self.DrawRubberBand(self.start_x, self.end_x, self.start_y, self.end_y)

    def InitializeScreenDrawing(self):
        # make an image of the currently rendered image

        rwi = self.GetInteractor()
        rwin = rwi.GetRenderWindow()

        size = rwin.GetSize()

        self._pixel_array.Initialize()
        self._pixel_array.SetNumberOfComponents(4)
        self._pixel_array.SetNumberOfTuples(size[0] * size[1])

        front = 1  # what does this do?
        rwin.GetRGBACharPixelData(0, 0, size[0] - 1, size[1] - 1, front, self._pixel_array)

    def DrawRubberBand(self, x1, x2, y1, y2):
        rwi = self.GetInteractor()
        rwin = rwi.GetRenderWindow()

        size = rwin.GetSize()

        tempPA = vtk.vtkUnsignedCharArray()
        tempPA.DeepCopy(self._pixel_array)

        # check size, viewport may have been resized in the mean-time
        if tempPA.GetNumberOfTuples() != size[0] * size[1]:
            # print(
            #     "Starting new screen-image - viewport has resized without us knowing"
            # )
            self.InitializeScreenDrawing()
            self.DrawRubberBand(x1, x2, y1, y2)
            return

        x2 = min(x2, size[0] - 1)
        y2 = min(y2, size[1] - 1)

        x2 = max(x2, 0)
        y2 = max(y2, 0)

        # Modify the pixel array
        width = abs(x2 - x1)
        height = abs(y2 - y1)
        minx = min(x2, x1)
        miny = min(y2, y1)

        # draw top and bottom
        for i in range(width):

            # c = round((10*i % 254)/254) * 254  # find some alternating color
            c = 0

            idx = (miny * size[0]) + minx + i
            tempPA.SetTuple(idx, (c, c, c, 1))

            idx = ((miny + height) * size[0]) + minx + i
            tempPA.SetTuple(idx, (c, c, c, 1))

        # draw left and right
        for i in range(height):
            # c = round((10 * i % 254) / 254) * 254  # find some alternating color
            c = 0

            idx = ((miny + i) * size[0]) + minx
            tempPA.SetTuple(idx, (c, c, c, 1))

            idx = idx + width
            tempPA.SetTuple(idx, (c, c, c, 1))

        # and Copy back to the window
        rwin.SetRGBACharPixelData(0, 0, size[0] - 1, size[1] - 1, tempPA, 0)
        rwin.Frame()

    def LineToPixels(self, x1, x2, y1, y2):
        """Returns the x and y values of the pixels on a line between x1,y1 and x2,y2.
        If start and end are identical then a single point is returned"""

        dx = x2 - x1
        dy = y2 - y1

        if dx == 0 and dy == 0:
            return [x1], [y1]

        if abs(dx) > abs(dy):
            dhdw = dy / dx
            r = range(0, dx, int(dx / abs(dx)))
            x = [x1 + i for i in r]
            y = [round(y1 + dhdw * i) for i in r]
        else:
            dwdh = dx / dy
            r = range(0, dy, int(dy / abs(dy)))
            y = [y1 + i for i in r]
            x = [round(x1 + i * dwdh) for i in r]

        return x, y

    def DrawLine(self, x1, x2, y1, y2):
        rwi = self.GetInteractor()
        rwin = rwi.GetRenderWindow()

        size = rwin.GetSize()

        x1 = min(max(x1, 0), size[0])
        x2 = min(max(x2, 0), size[0])
        y1 = min(max(y1, 0), size[1])
        y2 = min(max(y2, 0), size[1])

        tempPA = vtk.vtkUnsignedCharArray()
        tempPA.DeepCopy(self._pixel_array)

        xs, ys = self.LineToPixels(x1, x2, y1, y2)
        for x, y in zip(xs, ys):
            idx = (y * size[0]) + x
            tempPA.SetTuple(idx, (0, 0, 0, 1))

        # and Copy back to the window
        rwin.SetRGBACharPixelData(0, 0, size[0] - 1, size[1] - 1, tempPA, 0)

        camera = self.GetCurrentRenderer().GetActiveCamera()
        scale = camera.GetParallelScale()

        # Set/Get the scaling used for a parallel projection, i.e.
        #
        # the half of the height of the viewport in world-coordinate distances.
        # The default is 1. Note that the "scale" parameter works as an "inverse scale"
        #  larger numbers produce smaller images.
        # This method has no effect in perspective projection mode

        half_height = size[1] / 2
        # half_height [px] = scale [world-coordinates]

        length = ((x2 - x1) ** 2 + (y2 - y1) ** 2) ** 0.5
        meters_per_pixel = scale / half_height
        meters = length * meters_per_pixel

        if camera.GetParallelProjection():
            print(f"Line length = {length} px = {meters} m")
        else:
            print("Need to be in non-perspective mode to measure. Press 2 or 3 to get there")

        if self.callbackMeasure:
            self.callbackMeasure(meters)

        #
        # # can we add something to the window here?
        # freeType = vtk.vtkFreeTypeTools.GetInstance()
        # textProperty = vtk.vtkTextProperty()
        # textProperty.SetJustificationToLeft()
        # textProperty.SetFontSize(24)
        # textProperty.SetOrientation(25)
        #
        # textImage = vtk.vtkImageData()
        # freeType.RenderString(textProperty, "a somewhat longer text", 72, textImage)
        # # this does not give an error, assume it works
        # #
        # textImage.GetDimensions()
        # textImage.GetExtent()
        #
        # # # Now put the textImage in the RenderWindow
        # rwin.SetRGBACharPixelData(0, 0, size[0] - 1, size[1] - 1, textImage, 0)

        rwin.Frame()

    def UpdateMiddleMouseButtonLockActor(self):

        if self.middle_mouse_lock_actor is None:
            # create the actor
            # Create a text on the top-rightcenter
            textMapper = vtk.vtkTextMapper()
            textMapper.SetInput("Middle mouse lock [m or space] active")
            textProp = textMapper.GetTextProperty()
            textProp.SetFontSize(12)
            textProp.SetFontFamilyToTimes()
            textProp.BoldOff()
            textProp.ItalicOff()
            textProp.ShadowOff()
            textProp.SetVerticalJustificationToTop()
            textProp.SetJustificationToCentered()
            textProp.SetColor((0, 0, 0))

            self.middle_mouse_lock_actor = vtk.vtkActor2D()
            self.middle_mouse_lock_actor.SetMapper(textMapper)
            self.middle_mouse_lock_actor.GetPositionCoordinate().SetCoordinateSystemToNormalizedDisplay()
            self.middle_mouse_lock_actor.GetPositionCoordinate().SetValue(0.5, 0.98)

            self.GetCurrentRenderer().AddActor(self.middle_mouse_lock_actor)

        self.middle_mouse_lock_actor.SetVisibility(self.middle_mouse_lock)
        self.DoRender()

    def DoRender(self):
        self.GetInteractor().Render()



"""Demo to show how to solve the Tower of Hanoi"""
# Credits:
# https://github.com/gjbex/training-material/blob/master/Misc/Notebooks/hanoi.ipynb
# Creative Commons Zero v1.0 Universal licence

from vedo import Plotter, Cylinder, Box
from copy import deepcopy


class Hanoi:
    def __init__(self, nr_disks):
        self._nr_disks = nr_disks
        self._towers = [list(range(nr_disks, 0, -1)), list(), list()]

    @property
    def nr_disks(self):
        return self._nr_disks

    @property
    def nr_moves(self):
        return 2**self.nr_disks - 1

    @property
    def towers(self):
        return deepcopy(self._towers)

    def tower(self, n):
        return self._towers[n].copy()

    def move_disk(self, from_tower, to_tower):
        disk = self._towers[from_tower].pop()
        self._towers[to_tower].append(disk)
        return disk, from_tower, to_tower

    def move_disks(self, n, from_tower, to_tower):
        if n == 1:
            yield self.move_disk(from_tower, to_tower)
        else:
            helper = 3 - from_tower - to_tower
            yield from self.move_disks(n - 1, from_tower, helper)
            yield self.move_disk(from_tower, to_tower)
            yield from self.move_disks(n - 1, helper, to_tower)

    def moves(self):
        yield from self.move_disks(self.nr_disks, 0, 1)


nr_disks = 5
hanoi = Hanoi(nr_disks)

tower_states = list([hanoi.towers])
for _ in hanoi.moves():
    tower_states.append(hanoi.towers)

disks = { hanoi.nr_disks - i : Cylinder(r=0.2*(hanoi.nr_disks-i+1), c=i)
          for i in range(hanoi.nr_disks) }

plt = Plotter(interactive=False, size=(800, 600), bg='wheat', bg2='lb')
plt += list(disks.values())
plt += Box(pos=(3,0,-0.5), size=(12,4,0.1))
cam = dict(
    pos=(14.60, -20.56, 7.680),
    focal_point=(3.067, 0.5583, 1.910),
    viewup=(-0.1043, 0.2088, 0.9724),
)
plt.show(camera=cam)

for t in range(len(tower_states)):
    state = tower_states[t]
    for tower_nr in range(3):
        for i, disk in enumerate(state[tower_nr]):
            disks[disk].pos([3 * tower_nr, 0, i+0.5])
    plt.render()
plt.interactive().close()




'''Draw a z = f(x,y) surface specified as
a string or as a reference to an external function.
Red points indicate where the function does not exist!'''
from vedo import dataurl, sin, cos, log, show, Text2D
from vedo.pyplot import plot

doc = Text2D(__doc__, pos='bottom-left', c='darkgreen', font='Quikhand')

############################################################### REAL
def f(x, y):
    return sin(2*x*y) * cos(3*y)/2
f1 = plot(f, c='summer') # use a colormap

# red dots are shown where the function does not exist (y>x):
def f(x, y):
    return sin(3*x) * log(x-y)/3
f2 = plot(f, texture=dataurl+'textures/paper3.jpg')

# specify x and y ranges and z vertical limits:
def f(x, y):
    return log(x**2+y**2-1)
f3 = plot(
    f,
    xlim=[-2,2],
    ylim=[-1,8],
    zlim=[-1,None],
    texture=dataurl+'textures/paper1.jpg',
)

show([(f1, 'y = sin(2*x*y) * cos(3*y) /2', doc),
      (f2, 'y = sin(3*x)*log(x-y)/3'),
      (f3, 'y = log(x**2+y**2-1)'),
     ], N=3, sharecam=False,
).close()

############################################################## COMPLEX
comment = """Vertical axis shows the real part of complex z:
    z = sin(log(x:doty))
Color map the value of the imaginary part
(green=positive, purple=negative)"""

plt = plot(lambda x,y: sin(log(x*y))/25, mode='complex', bins=(51,51))

show(plt, comment, viewup='z').close()



"""A uniform distribution on a plane
is not uniform on a sphere"""
import numpy as np
from vedo.pyplot import histogram
from vedo import Plotter

phi = np.random.rand(1000)*np.pi*2
the = np.random.rand(1000)*np.pi

h = histogram(the, phi, mode='spheric').add_scalarbar()

plt = Plotter(axes=12).add_ambient_occlusion(0.05)
plt.show(h, __doc__, viewup='z').close()



"""For each point finds the 12 closest ones and fit a sphere.
Color points from the size of the sphere radius."""
from vedo import *
from vedo.pyplot import histogram

plt = Plotter()

msh = Mesh(dataurl+"cow.vtk").c("cyan7")

pts1, pts2, vals, cols = [], [], [], []

msh_points = msh.points()
for i in range(0, msh.npoints, 10):
    p = msh_points[i]
    pts = msh.closest_point(p, n=12)  # find the n-closest points to p
    sph = fit_sphere(pts)             # find the fitting sphere
    if sph is None:
        continue

    value = sph.radius * 10
    color = color_map(value, "jet", 0, 1)  # map value to a RGB color
    n = versor(p - sph.center)  # unit vector from sphere center to p
    vals.append(value)
    cols.append(color)
    pts1.append(p)
    pts2.append(p + n / 8)

plt += msh
plt += Points(pts1, c=cols)
plt += Lines(pts1, pts2, c="black")
plt += histogram(vals, xtitle='radius', xlim=[0,2]).as2d(pos="bottom-left")
plt += __doc__

plt.show().close()



"""Extract points on the boundary of a mesh.
Add an ID label to all vertices."""
from vedo import *

# Load a mesh from a URL, compute normals, and clean duplicate points
b = Mesh(dataurl+'290.vtk')
b.compute_normals().clean().linewidth(0.1)

# Get the point IDs on the boundary of the mesh
pids = b.boundaries(return_point_ids=True)

# Get the points corresponding to the boundary IDs
bpts = b.points()

# Create a Points object to represent the boundary points
pts = Points(bpts[pids], r=10, c='red5')

# Create a Label object for all the vertices in the mesh
labels = b.labels('id', scale=10).c('green2')

# Show the mesh, boundary points, vertex labels, and docstring 
show(b, pts, labels, __doc__, zoom=2).close()



"""A gyroscope sitting on a pedestal.
The analysis is in terms of Lagrangian mechanics.
The Lagrangian variables are polar angle theta,
azimuthal angle phi, and spin angle psi"""
# (adapted from http://www.glowscript.org)
from vedo import *

# ############################################################ parameters
dt = 3e-03  # time step
Lshaft = 1  # length of gyroscope shaft
M = 1  # mass of gyroscope (massless shaft)
R = 0.4  # radius of gyroscope rotor
theta = 1.3  # initial polar angle of shaft (from vertical)
psidot = -40  # spinning angular velocity (rad/s)
phidot = 0  # (try -1 and +1 to get first and second pattern)

# ############################################################
g, r = 9.81, Lshaft / 2
I3 = 1 / 2 * M * R ** 2  # moment of inertia, I, of gyroscope about its own axis
I1 = M * r ** 2 + 1 / 2 * I3  # I about a line through the support, perpendicular to axis
phi = psi = thetadot = 0
x = vector(theta, phi, psi)  # Lagrangian coordinates
v = vector(thetadot, phidot, psidot)

# ############################################################ the scene
plt = Plotter()
plt += __doc__

shaft = Cylinder([[0, 0, 0], [Lshaft, 0, 0]], r=0.03, c="dg")
rotor = Cylinder([[Lshaft / 2.2, 0, 0], [Lshaft / 1.8, 0, 0]], r=R).texture(dataurl+'textures/white.jpg')
base  = Sphere([0, 0, 0], c="dg", r=0.03)
tip   = Sphere([Lshaft, 0, 0], c="dg", r=0.03)
gyro = shaft + rotor + base + tip  # group relevant meshes into single one of type Assembly
plt += gyro  # add it to Plotter list

pedestal = Box([0, -0.63, 0], height=0.1, length=0.1, width=1).texture(dataurl+'textures/wood1.jpg')
pedbase = Box([0, -1.13, 0], height=0.5, length=0.5, width=0.05).texture(dataurl+'textures/wood1.jpg')
pedpin = Pyramid([0, -0.08, 0], axis=[0, 1, 0], s=0.05, height=0.12).texture(dataurl+'textures/wood1.jpg')
formulas = Picture(dataurl+"images/gyro_formulas.png").alpha(0.9)
formulas.scale(0.0035).pos(-1.4, -1.1, -1.1)
plt += [pedestal + pedbase + pedpin + formulas]

# ############################################################ the physics
def loop_func(event):
    global  t, v, x
    t += dt

    st, ct, sp, cp = sin(x[0]), cos(x[0]), sin(x[1]), cos(x[1])

    thetadot, phidot, psidot = v  # unpack
    atheta = st*ct * phidot**2 + (M*g*r*st - I3*(psidot + phidot*ct) * phidot*st)/I1
    aphi = I3/I1 * (psidot + phidot * ct) * thetadot/st - 2 * ct * thetadot * phidot/st
    apsi = phidot * thetadot * st - aphi * ct
    a = vector(atheta, aphi, apsi)

    v += a * dt  # update velocities
    x += v * dt  # update Lagrangian coordinates

    gaxis = (Lshaft + 0.03) * vector(st * sp, ct, st * cp)
    # set orientation along gaxis and rotate it around its axis by psidot*t degrees
    gyro.orientation(gaxis, rotation=psidot * t, rad=True)
    plt.add(Point(gaxis, r=3, c="red4")).render()

t = 0
plt.add_callback("timer", loop_func)
plt.timer_callback("start")
plt.show().close()



"""Mirror a mesh along one of the Cartesian axes"""
from vedo import dataurl, Mesh, show

myted1 = Mesh(dataurl+"teddy.vtk")

myted2 = myted1.clone(deep=False).mirror("y")
myted2.pos(0,3,0).c("green")
fp = myted2.flagpole("mirrored\nmesh").follow_camera()

show(myted1, myted2, fp, __doc__, viewup="z", bg2='ly', axes=9)





"""Markers set, analogous to matplotlib"""
from vedo import Plotter, Marker, Text3D

symbols = ['.','o','O', '0', 'p','*','h','D','d','v','^','>','<','s', 'x', 'a']

plt = Plotter(size=(1500,300))
plt += __doc__

for i,s in enumerate(symbols):
    plt += Marker(s, filled=True).x(i*0.6).backcolor('blue5')
    plt += Text3D(s, pos=[i*0.6,-0.6,0], s=0.12, literal=True, font="Calco")

plt.show(zoom='tight').close()



"""Select cells inside a point loop"""
from vedo import *

mesh = Mesh(dataurl + "dolfin_fine.vtk").lw(0.1)

pts = [
    [0.85382618, 0.1909104],
    [0.85585967, 0.8721275],
    [0.07500188, 0.8680605],
    [0.10143717, 0.0607675],
]

# Make a copy and cut it
cmesh = mesh.clone().cut_with_point_loop(
    pts, on="cells", include_boundary=False, invert=False,
)
cmesh.lw(0.1).c("tomato")

line = Line(pts, closed=True).lw(5).c("green3")

show([(mesh, line), (cmesh, line, __doc__)], N=2).close()



"""
Reconstruct a polygonal surface
from a point cloud:

 1. An object is loaded and
    noise is added to its vertices.
 2. The point cloud is smoothened
    with MLS (Moving Least Squares)
 3. Impose a minimum distance among points
 4. A triangular mesh is extracted from
    this set of sparse Points.
"""
from vedo import dataurl, printc, Plotter, Points, Mesh, Text2D


plt = Plotter(shape=(1,5))
plt.at(0).show(Text2D(__doc__, s=0.75, font='Theemim', bg='green5'))

# 1. load a mesh
mesh = Mesh(dataurl+"apple.ply").subdivide()
plt.at(1).show(mesh)

# Add noise
pts0 = Points(mesh, r=3).add_gaussian_noise(1)
plt.at(2).show(pts0)

# 2. Smooth the point cloud with MLS
pts1 = pts0.clone().smooth_mls_2d(f=0.8)
printc("Nr of points before cleaning nr. points:", pts1.npoints)

# 3. Impose a min distance among mesh points
pts1.subsample(0.005)
printc("             after  cleaning nr. points:", pts1.npoints)
plt.at(3).show(pts1)

# 4. Reconstruct a polygonal surface from the point cloud
reco = pts1.reconstruct_surface(dims=100, radius=0.2).c("gold")
plt.at(4).show(reco, axes=7, zoom=1.2)

plt.interactive().close()



from vedo import *
from time import time

def loop_func(event):
    msh.rotate_z(0.1)
    txt.text(f"time: {event.time - t0} sec")
    plt.render()

t0 = time()
msh = Cube()
txt = Text2D(bg='yellow', font="Calco")

plt = Plotter(axes=1)
plt.add_callback("timer", loop_func)
plt.timer_callback("start")
plt.show(msh, txt)
plt.close()



"""Create an empty Figure to be filled in a loop
Any 3D Mesh object can be added to the figure!"""
from vedo import *
from vedo.pyplot import plot, Figure

settings.default_font = "Cartoons123"
settings.palette = 2
settings.remember_last_figure_format = True

# Create an empty Figure and plot on it
fig = Figure(
    xlim=(0,12),
    ylim=(-1.5, 1.5),
    padding=0,    # no extra space
    aspect=16/9,  # desired aspect ratio
    xtitle="speed [mph]",
    grid=True,
    axes=dict(axes_linewidth=3, xyframe_line=3),
)

for i in range(2,11,2):
    x = np.linspace(0, 4*np.pi, 20)
    y = np.sin(x) * np.sin(x/12) * i/5

    fig += plot(x, y, '-0', c=i, splined=True)

fig += Arrow([5,-1], [8,-1], s=0.5, c='green3')

# Add any number of polygonal Meshes.
# Use insert() to preserve the object aspect ratio inside the Figure coord system:
mesh = Mesh(dataurl+'cessna.vtk').c('blue5').scale(0.25).pos(4, 0.5, 0.5)
circle = Circle([5,0.5,-0.1], r=0.5, c='orange5')
fig.insert(mesh, circle)

show(fig, __doc__, size=(800,700), zoom='tight').close()





"""Slice multiple datasets"""
from vedo import Plotter, Text2D, load, dataurl, ScalarBar3D

volumes = [dataurl+'vase.vti', dataurl+'embryo.slc', dataurl+'head.vti']
volumes = load(volumes)
cmaps = ['hot_r', 'gist_ncar_r', 'bone_r']

########################################################################
def initfunc(iren, data):

    vol = data.mode(1).c('k').alpha([0, 0, 0.15, 0, 0])
    txt = Text2D(data.filename[-20:], font='Calco')
    plt.at(iren).show(vol, vol.box(), txt)

    def func(widget, event):
        zs = int(widget.value)
        widget.title = f"z-slice = {zs}"
        msh = vol.zslice(zs)
        msh.cmap(cmaps[iren]).lighting("off")
        msh.name = "slice"
        sb = ScalarBar3D(msh, c='k')
        plt.renderer = widget.renderer  # make it the current renderer
        plt.remove("slice", "ScalarBar3D").add(msh, sb)

    return func  # this is the actual function returned!


########################################################################
plt = Plotter(shape=(1, len(volumes)), sharecam=False, bg2='lightcyan')

for iren, data in enumerate(volumes):
    plt.add_slider(
        initfunc(iren, data), #func
        0, data.dimensions()[2],
        value=0,
        show_value=False,
        pos=[(0.1,0.1), (0.25,0.1)],
    )

plt.interactive().close()



"""Probe a Volume with a line
and plot the intensity values"""
from vedo import dataurl, Volume, probe_line, show
from vedo.pyplot import plot

vol = Volume(dataurl+'embryo.slc')
vol.add_scalarbar3d('wild-type mouse embryo', c='k')

p1, p2 = (50,50,50), (200,200,200)
pl = probe_line(vol, p1, p2, res=100).linewidth(4)

xvals = pl.points()[:,0]
yvals = pl.pointdata[0] # get the probed values along the line

fig = plot(
    xvals, yvals,
    xtitle=" ", ytitle="voxel intensity",
    aspect=16/9,
    spline=True,
    lc="r",       # line color
    marker="*",   # marker style
    mc="dr",      # marker color
    ms=0.9,       # marker size
)
fig.shift(0,25,0)

show(vol, pl, fig, __doc__, axes=dict(xygrid=0, yzgrid=0)).close()



"""Generate the Mandelbrot set as a color-mapped Picture object"""
import numpy as np
from vedo import Picture, dataurl, show


def mandelbrot(h=400, w=400, maxit=20, r=2):
    # Returns an image of the Mandelbrot fractal of size (h,w)
    x = np.linspace(-2.5, 1.5, 4*h+1)
    y = np.linspace(-1.5, 1.5, 3*w+1)
    A, B = np.meshgrid(x, y)
    C = A + B*1j
    z = np.zeros_like(C)
    divtime = maxit + np.zeros(z.shape, dtype=int)
    for i in range(maxit):
        z = z**2 + C
        diverge = abs(z) > r                    # who is diverging
        div_now = diverge & (divtime == maxit)  # who is diverging now
        divtime[div_now] = i                    # note when
        z[diverge] = r                          # avoid diverging too much
    return divtime

pic = Picture(mandelbrot()).cmap("RdGy")
show(pic, __doc__, axes=1, size=[800,600], zoom=1.4).close()

# Also:
# Picture(dataurl+"images/dog.jpg").cmap("RdGy").show().close()



"""
Set a jpeg background image
on a separate rendering layer
"""
from vedo import *

# Create a plotter object with 4 subrenderers 
# and individual camera for each one
plt = Plotter(
    N=4,
    sharecam=False, # each subrenderer has its own camera
    bg=dataurl+"images/tropical.jpg", # set the background image
)

# Load a 3D model of a flamingo and rotate it so it is upright
a1 = load(dataurl+"flamingo.3ds").rotate_x(-90)

# Display a docstring on the second subrenderer
plt.at(2).show(__doc__)

# Zoom in on the background image to fill the window
plt.background_renderer.GetActiveCamera().Zoom(1.8)

# Display a logo on the first subrenderer
plt.at(0).show(VedoLogo(distance=2))

# Display the flamingo model on the fourth subrenderer
plt.at(3).show(a1)

# Allow the plot to be interacted with and then close it
plt.interactive().close()



"""Generate a polygonal Mesh from a contour line"""
from vedo import dataurl, load, Line, show
from vedo.pyplot import histogram

shapes = load(dataurl + "timecourse1d.npy")  # list of lines
shape = shapes[56].mirror().rotate_z(-90)
cmap = "RdYlBu"

msh = shape.generate_mesh()  # Generate the Mesh from the line
msh.smooth()                 # make the triangles more uniform
msh.compute_quality()        # add a measure of triangle quality
msh.cmap(cmap, on="cells").add_scalarbar3d()

contour = Line(shape).c("red4").lw(5)
labels = contour.labels("id")

histo = histogram(
    msh.celldata["Quality"], xtitle="triangle mesh quality", aspect=3/4, c=cmap,
)

show([(contour, labels, msh, __doc__), histo], N=2, sharecam=0).close()



"""Slice a Volume with an arbitrary plane
hover the plane to get the scalar values"""
from vedo import dataurl, precision, Sphere, Volume, Plotter

def func(evt):
    if not evt.actor:
        return
    pid = evt.actor.closest_point(evt.picked3d, return_point_id=True)
    txt = f"Probing:\n{precision(evt.actor.picked3d, 3)}\nvalue = {arr[pid]}"

    pts = evt.actor.points()
    sph = Sphere(pts[pid], c='orange7').pickable(False)
    fp = sph.flagpole(txt, s=7, offset=(-150,15), font=2).follow_camera()
    # remove old and add the two new objects
    plt.remove('Sphere', 'FlagPole').add(sph, fp).render()

vol = Volume(dataurl+'embryo.slc').alpha([0,0,0.8]).c('w').pickable(False)
vslice = vol.slice_plane(origin=vol.center(), normal=(0,1,1))
vslice.cmap('Purples_r').lighting('off').add_scalarbar('Slice', c='w')
arr = vslice.pointdata[0] # retrieve vertex array data

plt = Plotter(axes=9, bg='k', bg2='bb')
plt.add_callback('as my mouse moves please call', func) # be kind to vedo ;)
plt.show(vol, vslice, __doc__)
plt.close()



"""Share the same color and transparency mapping across different volumes"""
from vedo import Volume, Line, show
import numpy as np

arr = np.zeros(shape=(50,60,70))
for i in range(50):
    for j in range(60):
        for k in range(70):
            arr[i,j,k] = k

vol1 = Volume(arr   ).mode(1).cmap('jet', alpha=[0,1], vmin=0, vmax=80).add_scalarbar("vol1")
vol2 = Volume(arr+30).mode(1).cmap('jet', alpha=[0,1], vmin=0, vmax=80).add_scalarbar("vol2")

# or equivalently, to set transparency:
# vol1.alpha([0,1], vmin=0, vmax=70)

# can also manually build an extra scalarbar object to span the whole range:
sb = Line([50,0,0],[50,50,0]).cmap('jet',[0,70]).add_scalarbar3d("vol2", c='black').scalarbar

show([(vol1, __doc__), (vol2, sb)], N=2, axes=1, elevation=-25)



"""Use two sliders to change
color and transparency of a mesh"""
from vedo import Plotter, Mesh, dataurl


def slider1(widget, event):
    mesh.color(widget.value)

def slider2(widget, event):
    mesh.alpha(widget.value)


mesh = Mesh(dataurl+"magnolia.vtk").flat().lw(0.1)

plt = Plotter()
plt += [mesh, __doc__]

plt.add_slider(
    slider1,
    xmin=-9,
    xmax=9,
    value=0,
    pos="bottom-right",
    title="color number",
)

plt.add_slider(
    slider2,
    xmin=0.01,
    xmax=0.99,
    value=0.5,
    c="blue",
    pos="bottom-right-vertical",
    title="alpha value (opacity)",
)

plt.show().close()



"""Threshold the original TetMesh
with a scalar array"""
from vedo import *

settings.use_depth_peeling = True

tetm = TetMesh(dataurl+'limb_ugrid.vtk')
tetm.color('prism').alpha([0,1])

# Threshold the tetrahedral mesh for values in the range:
tetm.threshold(above=0.9, below=1)
tetm.add_scalarbar3d('chem_0  expression levels', c='k', italic=1)

show([(tetm,__doc__),
       tetm.tomesh(shrink=0.9),
     ], N=2, axes=1,
).close()



"""
Perform other simple mathematical operation between 3d Volumes.
Possible operations are: +, -, /, 1/x, sin, cos, exp, log, abs, **2, sqrt,
  min, max, atan, atan2, median, mag, dot, gradient, divergence, laplacian.
Alphas defines the opacity transfer function in the scalar range.
"""
from vedo import *
printc(__doc__)

plt = Plotter(N=6)

v0 = Volume(dataurl+'embryo.slc').c(0)
v0.add_scalarbar3d()
plt.at(0).show("original", v0)

v1 = v0.clone().operation("gradient").operation("mag")
v1.add_scalarbar3d()
# print(v1.pointdata.keys())
plt.at(1).show("gradient", v1)

v2 = v0.clone().operation("divergence").c(2)
v2.add_scalarbar3d()
plt.at(2).show("divergence", v2)

v3 = v0.clone().operation("laplacian").c(3)
v3.add_scalarbar3d()
plt.at(3).show("laplacian", v3)

v4 = v0.clone().operation("median").c(4)
v4.add_scalarbar3d()
plt.at(4).show("median", v4)

v5 = v0.clone().operation("dot", v0).c(7)
v5.add_scalarbar3d()
plt.at(5).show("dot(v0,v0)", v5, zoom=1.3)

plt.interactive().close()


############################################################# example application
#Start with creating a masked Volume then compute its gradient and probe 2 points
msh = Ellipsoid()

vol = msh.signed_distance(dims=(20, 20, 20))
vol.threshold(above=0.0, replace=0.0)  # replacing all values outside to 0
vol.c("blue").alpha([0.9, 0.0]).alpha_unit(0.1).add_scalarbar3d()

vgrad = vol.operation("gradient")
printc(vgrad.pointdata.keys(), c='g')

grd = vgrad.pointdata['ImageScalarsGradient']
pts = vol.points()  # coords as numpy array
arrs = Arrows(pts, pts + grd*0.1).lighting('off')

pts_probes = [[0.2,0.5,0.5], [0.2,0.3,0.4]]
vects = probe_points(vgrad, pts_probes).pointdata['ImageScalarsGradient']
arrs_pts_probe = Arrows(pts_probes, pts_probes+vects)

plt = Plotter(axes=1, N=2)
plt.at(0).show("A masked Volume", vol)
plt.at(1).show("..compute its gradient and probe 2 points", arrs, arrs_pts_probe)
plt.interactive().close()



"""Give a cartoony appearance to a 3D mesh"""
from vedo import dataurl, settings, Plotter, Mesh, Text2D

# Enable depth peeling for rendering transparency, 
# and set the number of antialiasing samples to 8
settings.use_depth_peeling = True
settings.multi_samples = 8

# Create an instance of the Plotter class, 
# which creates a default camera needed for silhouette()
plt = Plotter()

# Create a Text2D object to display the docstring at the bottom-center of the plot,
# using the Bongas font with a size of 2 and a dark green background
txt = Text2D(__doc__, pos="bottom-center", font="Bongas", s=2, bg="dg")

# Load a mesh of a human figure, disable lighting (reflections), 
# set the color to pink, and set the transparency to 0.5
man = Mesh(dataurl + "man.vtk").lighting("off").c("pink").alpha(0.5)

# Load a mesh of a teddy bear, scale and rotate it around the z-axis,
ted = Mesh(dataurl + "teddy.vtk").scale(0.4).rotate_z(-45).pos(-1, -1, -1)
ted.lighting("off").c("sienna").alpha(0.1)

# Show the meshes, the default silhouette of the teddy bear mesh
plt.show(
    txt,
    ted,
    man,
    ted.silhouette(),  # default silhouette of the teddy bear mesh
    man.silhouette(feature_angle=40).linewidth(3).color("dr"),
    bg="wheat",        # set the background color to wheat
    bg2="lb",          # set the color of the gradient to light blue
    elevation=-80,     # set the elevation angle of the camera
    zoom=1.2,          # zoom in on the plot
)

# Close the plot
plt.close()



import numpy as np
from dolfin import *
import vedo

set_log_level(30)


class AllBoundaries(SubDomain):
    def inside(self, x, on_boundary):
        return on_boundary and x[0]<-10

def solve_problem(mesh, mfunc, force):
    V = VectorFunctionSpace(mesh, "CG", 1)
    u = TrialFunction(V)
    v = TestFunction(V)
    displacement = Function(V)

    bc = [DirichletBC(V, Constant((0,0)), mfunc, 1)]

    F = Constant(force)
    E = Constant(5000)
    nu = Constant(0.3)
    mu = E / (2.0*(1+nu))
    lmbda = E * nu / ((1.0+nu) * (1-2*nu))
    sigma = 2.0 * mu * sym(grad(u)) + lmbda * tr( sym(grad(u)) ) * Identity(2)
    solve(inner(sigma, grad(v)) * dx == inner(F, v) * dx, displacement, bc)
    displacement.set_allow_extrapolation(True)
    return displacement

def update(mesh, displacement):
    new_mesh = Mesh(mesh)
    ALE.move(new_mesh, displacement)
    return new_mesh

def remesh(mesh):
    if isinstance(mesh, vedo.Mesh):
        vmesh = mesh
    else:
        vmesh = vedo.Mesh([mesh.coordinates(), mesh.cells()])
    bpts = vmesh.compute_normals(cells=True).boundaries().join(reset=1) #extract boundary
    vz = vmesh.celldata["Normals"][0][2] # check z component of normals at first point
    bpts.generate_mesh(invert=vz<0).write('tmpmesh.xml') #vedo REMESHING + smoothing
    return Mesh("tmpmesh.xml")

#################################################################################
N = 20         # number of iterations of stretching
do_remesh = 0  # grab the boundary and remesh the interior at each iteration

circle = vedo.Circle(r=50)
mesh = remesh(circle)
half_circle = circle.boundaries().cut_with_plane(origin=[-10,0,0], normal='-x').z(2)
half_circle.linewidth(5).c("red4")

plt = vedo.Plotter(N=N, size=(2250, 1300))

meshes = [mesh]
displacements = []
for i in range(N):
    mfunc = MeshFunction('size_t', mesh, 1, mesh.domains())
    mfunc.set_all(0)
    allb = AllBoundaries()
    allb.mark(mfunc, 1)

    F = np.array([4, 2*(i-N/2)/N])
    displacement = solve_problem(mesh, mfunc, F)
    new_mesh = update(mesh, displacement)

    mesh = remesh(new_mesh) if do_remesh else new_mesh
    meshes.append(mesh)
    displacements.append(displacement)

    varrow = vedo.Arrow2D([0,0], F*15).z(1).c("red4")
    vmesh = vedo.Mesh([mesh.coordinates(), mesh.cells()]).c("k4").lc('k5')
    plt.at(i).show(f"t={i}, F={F}", half_circle, vmesh, varrow, zoom=1.5)

plt.interactive().close()





"""Solve the wave equation using finite differences and the Euler method"""
import numpy as np
from scipy.ndimage import gaussian_filter
from vedo import Plotter, Grid, Text2D

N = 400      # grid resolution
A, B = 5, 4  # box sides
end = 5      # end time
nframes = 150

X, Y = np.mgrid[-A:A:N*1j, -B:B:N*1j]
dx = X[1,0] - X[0,0]
dt = 0.1 * dx
time = np.arange(0, end, dt)
m = int(len(time)/nframes)

# initial condition (a ring-like wave)
Z0 = np.ones_like(X)
Z0[X**2+Y**2 < 1] = 0
Z0[X**2+Y**2 > 2] = 0
Z0 = gaussian_filter(Z0, sigma=4)
Z1 = np.array(Z0)

grid = Grid(s=(X[:,0], Y[0])).linewidth(0).lighting('glossy')
txt = Text2D(font='Brachium', pos='bottom-left', bg='yellow5')

cam = dict(
    pos=(5.715, -10.54, 12.72),
    focal_point=(0.1380, -0.7437, -0.5408),
    viewup=(-0.2242, 0.7363, 0.6384),
    distance=17.40,
)

plt = Plotter(axes=1, size=(1000,700), interactive=False)
plt.show(grid, txt, __doc__, camera=cam)

for i in range(nframes):
    # iterate m times before showing the frame
    for _ in range(m):
        ZC = Z1.copy()
        Z1[1:N-1, 1:N-1] = (
            2*Z1[1:N-1, 1:N-1]
            - Z0[1:N-1, 1:N-1]
            + (dt/dx)**2
            * (  Z1[2:N,   1:N-1]
               + Z1[0:N-2, 1:N-1]
               + Z1[1:N-1, 0:N-2]
               + Z1[1:N-1, 2:N  ]
               - 4*Z1[1:N-1, 1:N-1] )
        )
        Z0[:] = ZC[:]

    wave = Z1.ravel()
    txt.text(f"frame: {i}/{nframes}, height_max = {wave.max()}")
    grid.cmap("Blues", wave, vmin=-2, vmax=2)
    newpts = grid.points()
    newpts[:,2] = wave
    grid.points(newpts)  # update the z component
    plt.render()

plt.interactive()
plt.close()



# Scatter plot of a gaussian distribution
# with varying color and point sizes
from vedo import Text2D, Plotter
from vedo.pyplot import plot
import numpy as np

n = 500
x = np.random.randn(n)
y = np.random.randn(n)

# Define what size must have each marker:
marker_sizes = np.sin(2*x)/8

# Define a (r,g,b) list of colors for each marker:
marker_cols = np.c_[np.cos(2*x), np.zeros(n), np.zeros(n)]


txt0 = Text2D("A scatter plot of a\n2D gaussian distribution")
fig0 = plot(
    x, y,
    lw=0,                      # no joining of lines
    marker="*",                # marker style
    xtitle="variable A",
    ytitle="variable B",
    grid=False,
)

txt1 = Text2D("marker size proportional to sin(2x) ")
fig1 = plot(
    x, y,
    lw=0,
    marker="*",
    ms=marker_sizes,           # VARIABLE marker sizes
    mc='purple5',              # same fixed color for markers
    grid=False,
)

txt2 = Text2D("marker size proportional to sin(2x)\nred level   proportional to cos(2x)")
fig2 = plot(
    x, y,  lw=0,
    marker=">",
    ms=marker_sizes,           # VARIABLE marker sizes
    mc=marker_cols,            # VARIABLE marker colors
    grid=False,
)

plt = Plotter(N=3, size=(1800,500))
plt.at(0).show(fig0, txt0)
plt.at(1).show(fig1, txt1)
plt.at(2).show(fig2, txt2, zoom=1.2)
plt.interactive().close()



"""Build a volume from a mesh where the
foreground voxels are set to 255 and the background voxels are 0"""
from vedo import Mesh, dataurl, Plotter

surf = Mesh(dataurl+"bunny.obj").normalize().wireframe()

vol = surf.binarize(spacing=(0.02,0.02,0.02))
vol.alpha([0,0.6]).c('blue')

iso = vol.isosurface().color("blue5")

plt = Plotter(N=2, axes=9)
plt.at(0).show(vol, surf, __doc__)
plt.at(1).show("..the volume is isosurfaced:", iso)
plt.interactive().close()



"""Create a Volume from a numpy array"""
import numpy as np
from vedo import Volume, show

data_matrix = np.zeros([70, 80, 90], dtype=np.uint8)
data_matrix[ 0:30,  0:30,  0:30] = 1
data_matrix[30:50, 30:60, 30:70] = 2
data_matrix[50:70, 60:80, 70:90] = 3

vol = Volume(data_matrix, c=['white','b','g','r'], mode=1)
vol.add_scalarbar3d()

# optionally mask some parts of the volume (needs mapper='gpu'):
# data_mask = np.zeros_like(data_matrix)
# data_mask[10:65, 10:65, 20:75] = 1
# vol.mask(data_mask)

show(vol, __doc__, axes=1).close()



"""A Volume can have multiple
scalars associated to each voxel"""
from vedo import dataurl, Volume, printc, show
import numpy as np

vol = Volume(dataurl+'vase.vti')
nx, ny, nz = vol.dimensions()
r0,r1 = vol.scalar_range()
vol.add_scalarbar3d('original voxel scalars')

# create a set of scalars and add it to the Volume
vol.pointdata["myscalars1"] = np.linspace(r0,r1, num=nx*ny*nz)

# create another set of scalars and add it to the Volume
vol.pointdata["myscalars2"] = np.random.randint(-100,+100, nx*ny*nz)

# make SLCImage scalars the active array (can set 0, to pick the first):
printc('Arrays in Volume are:\n', vol.pointdata.keys(), invert=True)
vol.pointdata.select("SLCImage")  # select the first data array as the active one

# Build the isosurface of the active scalars,
# but use testscals1 to colorize this isosurface, and then smooth it
iso1 = vol.isosurface().cmap('jet', 'myscalars1').smooth().lw(0.1)
iso1.add_scalarbar3d('myscalars1')

iso2 = vol.isosurface().cmap('viridis', 'myscalars2')
iso2.add_scalarbar3d('myscalars2')

show([(vol, __doc__),
       (iso1,"Colorize isosurface using\nmyscalars1"),
       (iso2,"Colorize isosurface using\nmyscalars2"),
     ], N=3, axes=1
).close()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tetralize a closed surface mesh
Click on the mesh and press ↓ or x to toggle a piece"""
from vedo import dataurl, Sphere, settings, Mesh, TessellatedBox, show

settings.use_depth_peeling = True

surf = Sphere(quads=True, res=15)
# surf = TessellatedBox()
# surf = Mesh(dataurl+'290_closed.vtk')
# surf = Mesh(dataurl+'bunny.obj', c='g3').fill_holes().cap().smooth()

tmesh = surf.tetralize(side=0.015, debug=True)
#tmesh.write('mytetmesh.vtk')  # save to disk!

# Assign an id to each tetrahedron to visualize regions
seeds = surf.clone().subsample(0.3)
cids = []
for p in tmesh.cell_centers():
	cid = seeds.closest_point(p, return_point_id=True)
	cids.append(cid)
tmesh.celldata["fragments"] = cids

pieces = []
for i in range(seeds.npoints):
	tc = tmesh.clone().threshold("fragments", above=i-0.1, below=i+0.1)
	mc = tc.tomesh(fill=True, shrink=0.95).color(i)
	pieces.append(mc)

show(__doc__, pieces, axes=1)



"""Record and playback camera movements and other events
\rightarrow Move the cube around, press 1, and finally press q"""
from vedo import Cube, Plotter

plt1 = Plotter(axes=1, interactive=0, title="recording window")
evts = plt1.show(Cube(), __doc__).record()
# print("Events:", evts) # a simple string (also saved as .vedo_recorded_events.log)

plt2 = Plotter(axes=1, interactive=0, title="playback window", pos=(1100,0))
plt2.show(Cube(), "...now playing!").play(evts).interactive().close()



"""Explore RGB and HSV color spaces"""
from vedo import Plotter, Cube, Text2D, precision, np
from vedo.colors import rgb2hex, rgb2hsv, hsv2rgb, get_color_name

def update(rgb, hsv):
    box.color(rgb)
    RGB = np.round(np.array(rgb)*255).astype(int)
    tx1.text(f"RGB: {precision(rgb,3)}\n     {RGB}\nHEX: {rgb2hex(rgb)}")
    tx2.text(f"HSV: {precision(hsv,3)}\n{get_color_name(rgb)}")

def funcRGB(w, e):
    r,g,b = slr.value, slg.value, slb.value
    h,s,v = rgb2hsv([r,g,b])
    slh.value = h
    sls.value = s
    slv.value = v
    update([r,g,b], [h,s,v])

def funcHSV(w, e):
    h,s,v = slh.value, sls.value, slv.value
    r,g,b = hsv2rgb([h,s,v])
    slr.value = r
    slg.value = g
    slb.value = b
    update([r,g,b], [h,s,v])

box = Cube().lw(2).color([1,0,0]).lighting("off")
tx1 = Text2D(font="Calco", s=1.5, pos="top-left",  bg="k5").text(__doc__)
tx2 = Text2D(font="Calco", s=1.5, pos="top-right", bg="k5")

plt = Plotter()

slr = plt.add_slider(funcRGB, 0,1, value=1, show_value=False, c="r3", pos=((0.05, 0.18),(0.4, 0.18)))
slg = plt.add_slider(funcRGB, 0,1, value=0, show_value=False, c="g3", pos=((0.05, 0.12),(0.4, 0.12)))
slb = plt.add_slider(funcRGB, 0,1, value=0, show_value=False, c="b3", pos=((0.05, 0.06),(0.4, 0.06)), title="RGB")

slh = plt.add_slider(funcHSV, 0,1, value=0, show_value=False, c="k1", pos=((0.6, 0.18),(0.95, 0.18)))
sls = plt.add_slider(funcHSV, 0,1, value=1, show_value=False, c="k1", pos=((0.6, 0.12),(0.95, 0.12)))
slv = plt.add_slider(funcHSV, 0,1, value=1, show_value=False, c="k1", pos=((0.6, 0.06),(0.95, 0.06)), title="HSV")

plt.show(box, tx1, tx2, viewup="z")



"""Colorize a large cloud of 1M points by passing
colors and transparencies in the format (R,G,B,A)"""
import time
from vedo import *

N = 1000000

pts = np.random.rand(N, 3)
RGB = pts * 255
Alpha = pts[:, 2] * 255
RGBA = np.c_[RGB, Alpha]  # concatenate

t0 = time.time()

# passing c in format (R,G,B,A) is ~50x faster
pts = Points(pts, r=2, c=RGBA)

t1 = time.time()
print("-> elapsed time:", t1-t0, "seconds for N:", N)

show(pts, __doc__, axes=True).close()



"""Histogram of 2 variables as 3D bars"""
import numpy as np
from vedo import Points, show
from vedo.pyplot import histogram

n = 1000
x = np.random.randn(n)*1.5 + 60
y = np.random.randn(n)     + 70

histo = histogram(
    x, y,
    bins=(12, 10),
    cmap="summer",
    ztitle="Number of entries in bin",
    mode="3d",
    gap=0.0,
    zscale=0.4,  # rescale the z axis
    aspect=16/9,
)

print(histo.frequencies)

# Add also the original points on top
histo += Points(np.c_[x, y], r=4, c="red5").z(3)

show(histo, __doc__, elevation=-80).close()



from vedo import *

# mode = 1 is maximum projection (default is 0=composite)
v1 = Volume(dataurl+'embryo.tif').mode(1)
v1.add_scalarbar3d(c='w').print_histogram(logscale=1, horizontal=1, c='g')
t1 = Text2D('Original volume', c='lg')

# cutoff range is roughly in the range of 1 / size of object
v2 = v1.clone().frequency_pass_filter(high_cutoff=.001, order=1).mode(1)
v2.add_scalarbar3d(c='w').print_histogram(logscale=1, horizontal=1, c='b')
t2 = Text2D('High freqs in the FFT\nare cut off:', c='lb')

show([(v1,t1), (v2,t2)], N=2, bg='bb', zoom=1.5, axes=dict(digits=2)).close()
#write(v2, 'embryo_filtered.vti')



from vedo import dataurl, Volume
from vedo.applications import IsosurfaceBrowser

vol = Volume(dataurl+'head.vti')

# IsosurfaceBrowser(Plotter) instance:
plt = IsosurfaceBrowser(vol, use_gpu=True, c='gold')

plt.show(axes=7, bg2='lb').close()



"""Generate 3 random sets of points
and align them using Procrustes method"""
from random import uniform as u
from vedo import Plotter, procrustes_alignment, Points

# Define number of points and a randomness factor
N = 15  # number of points
x = 1.0  # add some randomness

# Generate 3 sets of random points
pts1 = [(u(0, x), u(0, x), u(0, x) + i) for i in range(N)]
pts2 = [(u(0, x) + 3, u(0, x) + i / 2 + 2, u(0, x) + i + 1) for i in range(N)]
pts3 = [(u(0, x) + 4, u(0, x) + i / 4 - 3, u(0, x) + i - 2) for i in range(N)]

# Convert the sets of points into Points objects with different colors and sizes
vpts1 = Points(pts1, c="r", r=8)
vpts2 = Points(pts2, c="g", r=8)
vpts3 = Points(pts3, c="b", r=8)

# Perform Procrustes alignment on the sets of points
#  and obtain the aligned sets
# return an Assembly object formed by the aligned sets
aligned = procrustes_alignment([vpts1, vpts2, vpts3])
#print([aligned.transform])

# Create a Plotter object with a 1x2 grid, 2D axes, 
#  and independent camera control
plt = Plotter(shape=[1,2], axes=2, sharecam=False)
plt.at(0).show(vpts1, vpts2, vpts3, __doc__)
plt.at(1).show(aligned)
plt.interactive().close()



"""Setting illumination properties:
ambient, diffuse, specular, specularPower, specularColor.
"""
from vedo import Plotter, Mesh, dataurl


plt = Plotter(axes=1)

ambient, diffuse, specular = 0.1, 0., 0.
specularPower, specularColor= 20, 'white'

apple = Mesh(dataurl+'apple.ply').normalize().c('gold')

for i in range(8):
    s = apple.clone().pos((i%4)*2.2, int(i<4)*3, 0)

    #s.phong()
    s.flat()

    # modify the default with specific values
    s.lighting('default', ambient, diffuse, specular, specularPower, specularColor)
    #ambient += 0.125
    diffuse += 0.125
    specular += 0.125

    plt += s

plt += __doc__
plt.show().close()



"""pygeodesic library to compute geodesic distances"""
from pygeodesic import geodesic  # pip install pygeodesic
import vedo

m = vedo.Mesh(vedo.dataurl+"bunny.obj").c("green9")

geoalg = geodesic.PyGeodesicAlgorithmExact(m.points(), m.faces())

# Use source and target point ids
distance, path = geoalg.geodesicDistance(639, 834)
distances, _   = geoalg.geodesicDistances([639, 1301]) # any of the two

line = vedo.Line(path).c("k").lw(4)
m.cmap("Set2", distances, name="GeodesicDistance")

vedo.show(m, line, __doc__, axes=1)




"""Customizing a legend box"""
from vedo import *

s = Sphere()
c = Cube().x(2)
e = Ellipsoid().x(4)
h = Hyperboloid().x(6).legend('The description for\nthis one is quite long')

lb = LegendBox([s,c,e,h], width=0.3, height=0.4, markers='s').font("Kanopus")

cam = dict( # press C in window to get these numbers
    position=(10.4414, -7.62994, 4.18818),
    focal_point=(4.10196, 0.335224, -0.148651),
    viewup=(-0.252830, 0.299657, 0.919936),
    distance=11.0653,
    clipping_range=(3.69605, 21.2641),
)

show(s, c, e, h, lb, __doc__,
     axes=1, bg='lightyellow', bg2='white', size=(1400,800), camera=cam
).close()




"""Create a range slider to scale two spheres"""
from vedo import *


def slider1(w, e):
    if slid1.value > slid2.value:
        slid1.value = slid2.value
    s1.scale(slid1.value, reset=True)


def slider2(w, e):
    if slid2.value < slid1.value:
        slid2.value = slid1.value
    s2.scale(slid2.value, reset=True)


s1 = Sphere().c("red5").alpha(0.5).scale(0.8)
s2 = Sphere().c("green4").alpha(0.5).scale(1.2)

plt = Plotter()

slid2 = plt.add_slider(
    slider2,
    xmin=0.1,
    xmax=2,
    value=1.2,
    slider_length=0.02,
    slider_width=0.06,
    alpha=0.5,
    c="green4",
    show_value=True,
    font="Calco",
)

slid1 = plt.add_slider(
    slider1,
    xmin=0.1,
    xmax=2.0,
    value=0.8,
    slider_length=0.01,
    slider_width=0.05,
    alpha=0.5,
    tube_width=0.0015,
    c="red5",
    show_value=True,
    font="Calco",
)

plt.show(s1, s2, __doc__, axes=1).close()



"""Multiple plotter sync-ed windows"""
from vedo import *

acts = [Ellipsoid().color('Bisque'),
        Cone().color('RosyBrown'),
        Cylinder().color('Chocolate'),
]

opts = dict(axes=1, interactive=False, new=True, size=(390,390))
ts = [f"Window nr.{i}" for i in range(4)]

plt0 = show(acts[0], **opts, pos=( 200,0), title=ts[0], viewup='z')
plt1 = show(acts[1], **opts, pos=( 600,0), title=ts[1], camera=plt0.camera)
plt2 = show(acts[2], __doc__, **opts, pos=(1000,0), title=ts[2], camera=plt0.camera)
plts = [plt0, plt1, plt2]

def func(evt):
    for i in range(3):
        if ts[i] != evt.title: # only update the other windows
            plts[i].render()

for plt in plts:
    plt.add_callback('Interaction', func)
    plt.add_callback('EndInteraction', func) # because zooming is not an "Interaction" event

plt.interactive()



from vedo import Latex, show
from vedo.pyplot import histogram
import numpy as np

N = 2000
x = np.random.randn(N) * 1.0
y = np.random.randn(N) * 1.5

# hexagonal binned histogram:
histo = histogram(
    x, y,
    bins=10,
    mode='hexbin',
    fill=True,
    cmap='terrain',
)
# add a formula:
f = r'f(x, y)=A \exp \left(-\left(\frac{\left(x-x_{o}\right)^{2}}'
f+= r'{2 \sigma_{x}^{2}}+\frac{\left(y-y_{o}\right)^{2}}'
f+= r'{2 \sigma_{y}^{2}}\right)\right)'
formula = Latex(f, c='k', s=1.5).rotate_x(90).rotate_z(90).pos(-4,-5,2)

show(histo, formula, axes=1, viewup='z')



"""Generate a Volume with the signed distance from a Mesh,
then generate the isosurface at distance -0.5"""
from vedo import *

mesh = Mesh(dataurl+"beethoven.ply").subdivide()
mesh.color('k').point_size(3) # render mesh as points

# Generate signed distance volume
vol = mesh.signed_distance(dims=(40,40,40))

# Generate an isosurface at distance -0.5
iso = vol.isosurface(-0.5)

show(mesh, iso, __doc__, axes=1).close()



"""Generate the silhouette of a mesh
as seen along a specified direction

Axes font: """
# Source: Zhi-Qiang Zhou (https://github.com/zhouzq-thu)
from vedo import *

settings.default_font = "Kanopus"
settings.use_depth_peeling = False

plt = Plotter(title="Example of project_on_plane()")

s = Hyperboloid().rotate_x(20)
pts = s.points()
n = len(pts)

plt += [s, __doc__ + settings.default_font]

# orthogonal projection ###############################
plane1 = Plane(pos=(2, 0, 2), normal=(1, 0, 1), s=[5, 5]).alpha(0.1)
so = s.clone().project_on_plane(plane1).c("y")
plt += [plane1, so, so.silhouette("2d")]
pts1 = so.silhouette("2d").points()

# perspective projection ##############################
plane2 = Plane(pos=(3, 3, 3), normal=(1, 1, 1), s=[5, 5]).alpha(0.1)
point = [6, 6, 6]
sp = s.clone().project_on_plane(plane2, point=point).c("m")
plt += [plane2, sp, sp.silhouette("2d")]

# oblique projection ##################################
plane3 = Plane(pos=(4, 8, -4), normal=(-1, 0, 1), s=[5, 5]).alpha(0.1)
sob = s.clone().project_on_plane(plane3, direction=(1, 2, -1)).c("g")
plt += [plane3, sob, sob.silhouette("2d")]
pts2 = sob.silhouette("2d").points()

# draw the lines
for i in range(0, n, int(n / 20)):
    plt += Line(pts1[i], pts[i], c="k", alpha=0.2)
    plt += Line(point, pts[i], c="k", alpha=0.2)
    plt += Line(pts2[i], pts[i], c="k", alpha=0.2)

plt.show(
    axes=dict(
        xtitle="X-axis in :mum",
        ytitle="Y-axis in :mum",
        ztitle="Z-axis in :mum",
        yzgrid=False,
        text_scale=1.5,
    ),
).close()



"""Cauchy-Green and Green-Lagrange strain tensors on a 2D grid."""
import numpy as np
import vedo

# Define a simple deformation function
def deform(x, y):
    xd = np.array([x + 0.25 * x * x, 
                   y + 0.25 * x * y])
    # Add rotation to the deformation.
    # note that the rotation is applied to the deformed configuration
    # and it has no effect on the deformation gradient tensor C and E
    rotation_angle_degrees = 10
    rotation_angle_radians = np.radians(rotation_angle_degrees)
    cos_angle = np.cos(rotation_angle_radians)
    sin_angle = np.sin(rotation_angle_radians)
    x_def, y_def = xd
    x_rot = x_def * cos_angle - y_def * sin_angle
    y_rot = x_def * sin_angle + y_def * cos_angle
    return np.array([x_rot, y_rot])

# Compute the deformation gradient tensor F
def deformation_gradient(x, y, ds=0.001):
    # Compute the deformation gradient tensor F
    # F = (df/dx, df/dy)
    fxy = deform(x, y)
    fxy_x = deform(x + ds, y)
    fxy_y = deform(x, y + ds)
    F = np.zeros((2, 2))
    F[0, 0] = (fxy_x[0] - fxy[0]) / ds
    F[0, 1] = (fxy_y[0] - fxy[0]) / ds
    F[1, 0] = (fxy_x[1] - fxy[1]) / ds
    F[1, 1] = (fxy_y[1] - fxy[1]) / ds
    return F

# Compute the right Cauchy-Green deformation tensor C
def cauchy_green(F):
    return F.T @ F

# Right Cauchy-Green tensor (C) is used to define the Green-Lagrange
# strain tensor (E), which is a measure of deformation in the reference
# (undeformed) configuration:
def green_lagrange(C):
    return 0.5 * (C - np.eye(2))

# Left Cauchy-Green tensor (B) is used to define the Almansi strain tensor (e),
# which is a measure of deformation in the current (deformed) configuration:
# e = 0.5 * (I - B^-1) (this is less used in practice)
def almansi(F):
    B = F @ F.T
    return 0.5 * (np.eye(2) - np.linalg.inv(B))

# Compute the principal stretches and directions
def principal_stretches_directions(T):
    # T is a symmetric tensor
    # eigenvalues are sorted in ascending order
    eigenvalues, eigenvectors = np.linalg.eigh(T)
    principal_stretches = np.sqrt(np.abs(eigenvalues))*np.sign(eigenvalues)
    return principal_stretches, eigenvectors

######################################################################
# Define the original grid (undeformed configuration)
x, y = np.meshgrid(np.linspace(-1, 1, 8), np.linspace(-1, 1, 8))

grid = vedo.Grid(s=(x[0], y.T[0]))
grid_pts = grid.points()
grid_pts_defo = deform(grid_pts[:, 0], grid_pts[:, 1])
grid_defo = grid.clone().points(grid_pts_defo.T)

# Initialize the vedo plotter
plotter = vedo.Plotter()

for i in range(x.shape[0]):
    for j in range(y.shape[1]):
        pt = x[i, j], y[i, j]
        displaced_pt = deform(*pt)
        F = deformation_gradient(*pt)

        C = cauchy_green(F)
        stretches, directions = principal_stretches_directions(C)
        ellipsoid_axes = np.diag(stretches) @ directions.T / 8
        ellipsoid_C = vedo.Ellipsoid(
            axis1=ellipsoid_axes[0],
            axis2=ellipsoid_axes[1],
            axis3=[0, 0, 0.01],
            pos=(*pt, 0),
            c="blue5",
        ).lighting("off")

        E = green_lagrange(C)
        # E = almansi(F)
        stretches, directions = principal_stretches_directions(E)
        ellipsoid_axes = np.diag(stretches) @ directions.T / 8
        ellipsoid_E = vedo.Ellipsoid(
            axis1=ellipsoid_axes[0],
            axis2=ellipsoid_axes[1],
            axis3=[0, 0, 0.01],
            pos=(*pt, 0),
            c="purple5",
        ).z(0.01).lighting("off")
        if stretches[0] < 0 or stretches[1] < 0:
            ellipsoid_E.c("red4")

        # Plot the deformation gradient tensor, we cannot compute the
        # principal stretches and directions of the deformation gradient
        # tensor because it is not a symmetric tensor.
        # F = deformation_gradient(*pt)
        # circle = vedo.Circle(r=0.05, c="black").pos(*pt)
        # cpts = circle.points()
        # cpts_defo = F @ cpts.T[:2]
        # circle.points(cpts_defo.T)
        # Same as:
        circle = vedo.Circle(r=0.06, c="black").pos(*pt)
        cpts = circle.points()
        cpts_defo = deform(cpts[:,0], cpts[:,1])
        circle.points(cpts_defo.T)

        plotter += [ellipsoid_C, ellipsoid_E, circle]

pts  = np.array([x, y]).T.reshape(-1, 2)
defo_pts = deform(x, y).T.reshape(-1, 2)

plotter += vedo.Arrows2D(pts, defo_pts, s=0.2, c="blue5")
plotter += grid_defo
plotter += __doc__
plotter.show(axes=8, zoom=1.2)

#####################################################################
# Resources:
# https://en.wikipedia.org/wiki/Deformation_gradient
# https://en.wikipedia.org/wiki/Almansi_strain_tensor
# https://en.wikipedia.org/wiki/Principal_stretch
# https://www.continuummechanics.org/deformationstrainintro.html


"""Insert a Figure into another (note that the x-axes stay aligned)"""
from vedo import Marker, settings, show, np
from vedo.pyplot import histogram

settings.default_font = "Ubuntu"

data = np.random.normal(loc=100, size=1000) + 7

################## Create the first Figure
fig1 = histogram(
    data,
    bins=20,
    xlim=(95,111),
    aspect=16/9,
    xtitle="shifted gaussian",
    c='cyan3',
)
# let's add an asterix marker where the mean is
fig1 += Marker('a', [fig1.mean,150,0.1], s=8).c('orange5')

################## Create a second Figure
fig2 = histogram(
    data - 7,
    bins=60,
    aspect=4/3,
    density=True,
    outline=True,
    c='purple9',
    axes=dict(xygrid=True, xyplane_color='grey2', xyalpha=1, grid_linewidth=0),
    label='finer binning',
)
# let's add an asterix marker where the mean is
fig2 += Marker('a', [fig2.mean,0.2,0.1], s=0.02).c('orange5')

# shift fig2 in vertical by 25, and in z by 0.1 (to make it show on top)
fig2.shift(0, 25, 0.1)

################## Insert fig2 into fig1
fig2.add_legend()
fig1.insert(fig2)

show(fig1, __doc__, zoom='tight', size=(1200,900)).close()




"""Segment a TetMesh with a custom scalar.
Press q to make it explode"""
from vedo import TetMesh, Plotter, Text2D, dataurl

n = 20000
f1 = 0.005  # control the tetras resolution
f2 = 0.15   # control the nr of seeds

tmesh = TetMesh(dataurl + "limb_ugrid.vtk")
surf = tmesh.tomesh(fill=False)
txt = Text2D(__doc__, font="Brachium")

# pick points on the surface and use subsample to make them uniform
seeds = surf.clone().subsample(f2).ps(10).c("black")

# assign to each tetrahedron the id of the closest seed point
cids = []
for p in tmesh.cell_centers():
    cid = seeds.closest_point(p, return_point_id=True)
    cids.append(cid)
tmesh.celldata["fragment"] = cids
# tmesh.celldata.select("fragment")# bug, has no effect, needs name=...

pieces = []
for i in range(seeds.npoints):
    tc = tmesh.clone().threshold(name="fragment", above=i-0.1, below=i+0.1)
    mc = tc.tomesh(fill=False).color(i)
    pieces.append(mc)

############### animate
plt = Plotter(size=(1200, 800), axes=1)
plt.show(txt, pieces)
for i in range(20):
    for pc in pieces:
        cm = pc.center_of_mass()
        pc.shift(cm / 25)
    txt.text(f"{__doc__}\n\nNr. of pieces = {seeds.npoints}")
    plt.render()
plt.interactive().close()



"""Create a simple Play/Pause app with a timer event
You can interact with the scene during the loop!
..press q to quit"""
import time
import numpy as np
from vedo import Plotter
from vedo.pyplot import plot


def bfunc():
    global timer_id
    plotter.timer_callback("destroy", timer_id)
    if "Play" in button.status():
        # instruct to call handle_timer() every 10 msec:
        timer_id = plotter.timer_callback("create", dt=10)
    button.switch()

def handle_timer(event):
    t = time.time() - t0
    x = np.linspace(t, t + 4*np.pi, 50)
    y = np.sin(x) * np.sin(x/12)
    fig = plot(
        x, y, '-o', ylim=(-1.2, 1.2), aspect=3/1,
        xtitle="time window [s]", ytitle="intensity [a.u.]",
    )
    fig.shift(-x[0]) # put the whole plot object back at (0,0)
    # pop (remove) the old plot and add the new one
    plotter.pop().add(fig).render()


timer_id = -1
t0 = time.time()
plotter= Plotter(size=(1200,600))
button = plotter.add_button(bfunc, states=[" Play ","Pause"], size=40)
evntId = plotter.add_callback("timer", handle_timer)

x = np.linspace(0, 4*np.pi, 50)
y = np.sin(x) * np.sin(x/12)
fig = plot(x, y, ylim=(-1.2, 1.2), xtitle="time", aspect=3/1, lc='grey5')

plotter.show(__doc__, fig, zoom=2)



"""Generate a Volume by interpolating a scalar
which is only known on a scattered set of points or mesh.
The blue layer is the result of thresholding the volume
between 0.3 and 0.4 and assigning it the new value 0.9 (blue)"""
from vedo import Points, show
from vedo.pyplot import CornerHistogram
import numpy as np

npts = 500                       # nr. of points where the scalar value is known
coords = np.random.rand(npts, 3) # range is [0, 1]
scals = coords[:, 2]             # let the scalar be the z of the point itself

pts = Points(coords)
pts.pointdata["scals"] = scals

# Now interpolate the values at these points to the full Volume
# available interpolation kernels are: shepard, gaussian, voronoi, linear.
vol = pts.tovolume(kernel='shepard', n=4, dims=(90,90,90))

vol.c(["maroon","g","b"])        # set color   transfer function
vol.alpha([0.3, 0.9])            # set opacity transfer function
#vol.alpha([(0.3,0.3), (0.9,0.9)]) # alternative way, by specifying (xscalar, alpha)
vol.alpha_unit(0.5)              # make the whole object less transparent (default is 1)

# replace voxels of specific range with a new value
vol.threshold(above=0.3, below=0.4, replace=0.9)
# Note that scalar range now has changed (you may want to reapply vol.c().alpha())

ch = CornerHistogram(vol, pos="bottom-left")

vol.add_scalarbar3d('Height is the voxel scalar', size=[None,1])
vol.scalarbar.rotate_x(90).pos(1.15,1,0.5)

show(pts, vol, ch, __doc__, axes=1, elevation=-90).close()



# 2D Fast Fourier Transform of a picture
from vedo import Picture, show

# url = 'https://comps.canstockphoto.com/a-capital-silhouette-stock-illustrations_csp31110154.jpg'
url = 'https://vedo.embl.es/examples/data/images/dog.jpg'

pic = Picture(url).resize([200,None])  # resize so that x has 200 pixels, but keep y aspect-ratio
picfft = pic.fft(logscale=12)
picfft = picfft.tomesh().cmap('Set1',"RGBA").add_scalarbar("12\dotlog(fft)")  # optional step

show([
      [pic, f"Original image\n{url[-40:]}"],
      [picfft, "2D Fast Fourier Transform"],
      [pic.fft(mode='complex').rfft(), "Reversed FFT"],
     ], N=3, bg='gray7', axes=1,
).close()





"""Start typing a color name then press return
E.g. pink4"""
from vedo import settings, dataurl, Plotter, Mesh

settings.enable_default_keyboard_callbacks = False

def kfunc(evt):
    global msg
    evt.keypress = evt.keypress.replace("period", ".")
    if evt.keypress == "BackSpace" and msg:
        msg = msg[:-1]
        evt.keypress = ''
    elif evt.keypress == "Return":
        bfunc()
        return
    elif evt.keypress == "Escape":
        plt.close()

    if len(evt.keypress) > 1:
        return

    msg += f"{evt.keypress}"
    bu.actor.SetInput(msg)
    plt.render()

def bfunc():
    mesh.color(msg)
    plt.render()


plt = Plotter(axes=1)
plt.interactor.RemoveObservers("CharEvent") # might be needed

msg = ""
plt.add_callback("key press", kfunc)

bu = plt.add_button(
    bfunc,
    pos=(0.7, 0.05),  # x,y fraction from bottom left corner
    states=["input box"],
    c=["w"],
    bc=["dg"],        # colors of states
    font="courier",   # arial, courier, times
    size=45,
    bold=True,
)

mesh = Mesh(dataurl+"magnolia.vtk").c("v").flat()

plt.show(mesh, __doc__).close()




"""Dijkstra algorithm to compute the graph geodesic.

Takes as input a polygonal mesh and performs
a shortest path calculation 20 times"""
from vedo import Sphere, Earth, show

msh = Sphere(r=1.02, res=200).subsample(0.007).wireframe().alpha(0.1)
# msh.triangulate().clean() # often needed!

path = msh.geodesic([0.349,-0.440,0.852], [-0.176,-0.962,0.302]).c("red4")
# path = msh.geodesic(10728, 9056).c("red4") # use vertex indices

# printc(geo.pointdata["VertexIDs"])

show(Earth(), msh, __doc__, path, bg2='lb', viewup="z", zoom=1.3).close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

try:
    import vedo.vtkclasses as vtk
except ImportError:
    import vtkmodules.all as vtk

import numpy as np
import vedo
from vedo import utils
from vedo.base import BaseGrid
from vedo.mesh import Mesh
from vedo.file_io import download, loadUnStructuredGrid


__docformat__ = "google"

__doc__ = """
Work with tetrahedral meshes.

![](https://vedo.embl.es/images/volumetric/82767107-2631d500-9e25-11ea-967c-42558f98f721.jpg)
"""

__all__ = ["TetMesh", "delaunay3d"]


##########################################################################
def delaunay3d(mesh, radius=0, tol=None):
    """
    Create 3D Delaunay triangulation of input points.

    Arguments:

        radius : (float)
            specify distance (or "alpha") value to control output.
            For a non-zero values, only tetra contained within the circumsphere
            will be output.

        tol : (float)
            Specify a tolerance to control discarding of closely spaced points.
            This tolerance is specified as a fraction of the diagonal length of
            the bounding box of the points.
    """
    deln = vtk.vtkDelaunay3D()
    if utils.is_sequence(mesh):
        pd = vtk.vtkPolyData()
        vpts = vtk.vtkPoints()
        vpts.SetData(utils.numpy2vtk(mesh, dtype=np.float32))
        pd.SetPoints(vpts)
        deln.SetInputData(pd)
    else:
        deln.SetInputData(mesh.GetMapper().GetInput())
    deln.SetAlpha(radius)
    deln.AlphaTetsOn()
    deln.AlphaTrisOff()
    deln.AlphaLinesOff()
    deln.AlphaVertsOff()
    deln.BoundingTriangulationOff()
    if tol:
        deln.SetTolerance(tol)
    deln.Update()
    m = TetMesh(deln.GetOutput())
    m.pipeline = utils.OperationNode(
        "delaunay3d", c="#e9c46a:#edabab", parents=[mesh],
    )
    return m


def _buildtetugrid(points, cells):
    ug = vtk.vtkUnstructuredGrid()

    if len(points) == 0:
        return ug
    if not utils.is_sequence(points[0]):
        return ug

    if len(cells) == 0:
        return ug

    if not utils.is_sequence(cells[0]):
        tets = []
        nf = cells[0] + 1
        for i, cl in enumerate(cells):
            if i in (nf, 0):
                k = i + 1
                nf = cl + k
                cell = [cells[j + k] for j in range(cl)]
                tets.append(cell)
        cells = tets

    source_points = vtk.vtkPoints()
    varr = utils.numpy2vtk(points, dtype=np.float32)
    source_points.SetData(varr)
    ug.SetPoints(source_points)

    source_tets = vtk.vtkCellArray()
    for f in cells:
        ele = vtk.vtkTetra()
        pid = ele.GetPointIds()
        for i, fi in enumerate(f):
            pid.SetId(i, fi)
        source_tets.InsertNextCell(ele)
    ug.SetCells(vtk.VTK_TETRA, source_tets)
    return ug


##########################################################################
class TetMesh(BaseGrid, vtk.vtkVolume):
    """The class describing tetrahedral meshes."""

    def __init__(
        self,
        inputobj=None,
        c=("r", "y", "lg", "lb", "b"),  # ('b','lb','lg','y','r')
        alpha=(0.5, 1),
        alpha_unit=1,
        mapper="tetra",
    ):
        """
        Arguments:
            inputobj : (vtkDataSet, list, str)
                list of points and tet indices, or filename
            alpha_unit : (float)
                opacity scale
            mapper : (str)
                choose a visualization style in `['tetra', 'raycast', 'zsweep']`
        """
        vtk.vtkVolume.__init__(self)
        BaseGrid.__init__(self)

        self.useArray = 0

        # inputtype = str(type(inputobj))
        # print('TetMesh inputtype', inputtype)

        ###################
        if inputobj is None:
            self._data = vtk.vtkUnstructuredGrid()

        elif isinstance(inputobj, vtk.vtkUnstructuredGrid):
            self._data = inputobj

        elif isinstance(inputobj, vtk.vtkRectilinearGrid):
            r2t = vtk.vtkRectilinearGridToTetrahedra()
            r2t.SetInputData(inputobj)
            r2t.RememberVoxelIdOn()
            r2t.SetTetraPerCellTo6()
            r2t.Update()
            self._data = r2t.GetOutput()

        elif isinstance(inputobj, vtk.vtkDataSet):
            r2t = vtk.vtkDataSetTriangleFilter()
            r2t.SetInputData(inputobj)
            # r2t.TetrahedraOnlyOn()
            r2t.Update()
            self._data = r2t.GetOutput()

        elif isinstance(inputobj, str):
            if "https://" in inputobj:
                inputobj = download(inputobj, verbose=False)
            ug = loadUnStructuredGrid(inputobj)
            tt = vtk.vtkDataSetTriangleFilter()
            tt.SetInputData(ug)
            tt.SetTetrahedraOnly(True)
            tt.Update()
            self._data = tt.GetOutput()

        elif utils.is_sequence(inputobj):
            # if "ndarray" not in inputtype:
            #     inputobj = np.array(inputobj)
            self._data = _buildtetugrid(inputobj[0], inputobj[1])

        ###################
        if "tetra" in mapper:
            self._mapper = vtk.vtkProjectedTetrahedraMapper()
        elif "ray" in mapper:
            self._mapper = vtk.vtkUnstructuredGridVolumeRayCastMapper()
        elif "zs" in mapper:
            self._mapper = vtk.vtkUnstructuredGridVolumeZSweepMapper()
        elif isinstance(mapper, vtk.vtkMapper):
            self._mapper = mapper
        else:
            vedo.logger.error(f"Unknown mapper type {type(mapper)}")
            raise RuntimeError()

        self._mapper.SetInputData(self._data)
        self.SetMapper(self._mapper)
        self.color(c).alpha(alpha)
        if alpha_unit:
            self.GetProperty().SetScalarOpacityUnitDistance(alpha_unit)

        # remember stuff:
        self._color = c
        self._alpha = alpha
        self._alpha_unit = alpha_unit

        self.pipeline = utils.OperationNode(
            self, comment=f"#tets {self._data.GetNumberOfCells()}",
            c="#9e2a2b",
        )
        # -----------------------------------------------------------

    def _repr_html_(self):
        """
        HTML representation of the TetMesh object for Jupyter Notebooks.

        Returns:
            HTML text with the image and some properties.
        """
        import io
        import base64
        from PIL import Image

        library_name = "vedo.tetmesh.TetMesh"
        help_url = "https://vedo.embl.es/docs/vedo/tetmesh.html"

        arr = self.thumbnail()
        im = Image.fromarray(arr)
        buffered = io.BytesIO()
        im.save(buffered, format="PNG", quality=100)
        encoded = base64.b64encode(buffered.getvalue()).decode("utf-8")
        url = "data:image/png;base64," + encoded
        image = f"<img src='{url}'></img>"

        bounds = "<br/>".join(
            [
                utils.precision(min_x,4) + " ... " + utils.precision(max_x,4)
                for min_x, max_x in zip(self.bounds()[::2], self.bounds()[1::2])
            ]
        )

        help_text = ""
        if self.name:
            help_text += f"<b> {self.name}: &nbsp&nbsp</b>"
        help_text += '<b><a href="' + help_url + '" target="_blank">' + library_name + "</a></b>"
        if self.filename:
            dots = ""
            if len(self.filename) > 30:
                dots = "..."
            help_text += f"<br/><code><i>({dots}{self.filename[-30:]})</i></code>"

        pdata = ""
        if self._data.GetPointData().GetScalars():
            if self._data.GetPointData().GetScalars().GetName():
                name = self._data.GetPointData().GetScalars().GetName()
                pdata = "<tr><td><b> point data array </b></td><td>" + name + "</td></tr>"

        cdata = ""
        if self._data.GetCellData().GetScalars():
            if self._data.GetCellData().GetScalars().GetName():
                name = self._data.GetCellData().GetScalars().GetName()
                cdata = "<tr><td><b> cell data array </b></td><td>" + name + "</td></tr>"

        pts = self.points()
        cm = np.mean(pts, axis=0)

        allt = [
            "<table>",
            "<tr>",
            "<td>", image, "</td>",
            "<td style='text-align: center; vertical-align: center;'><br/>", help_text,
            "<table>",
            "<tr><td><b> bounds </b> <br/> (x/y/z) </td><td>" + str(bounds) + "</td></tr>",
            "<tr><td><b> center of mass </b></td><td>" + utils.precision(cm,3) + "</td></tr>",
            "<tr><td><b> nr. points&nbsp/&nbsptets </b></td><td>"
            + str(self.npoints) + "&nbsp/&nbsp" + str(self.ncells) + "</td></tr>",
            pdata,
            cdata,
            "</table>",
            "</table>",
        ]
        return "\n".join(allt)


    def _update(self, data):
        self._data = data
        self.mapper().SetInputData(data)
        self.mapper().Modified()
        return self

    def clone(self):
        """Clone the `TetMesh` object to yield an exact copy."""
        ugCopy = vtk.vtkUnstructuredGrid()
        ugCopy.DeepCopy(self._data)

        cloned = TetMesh(ugCopy)
        pr = vtk.vtkVolumeProperty()
        pr.DeepCopy(self.GetProperty())
        cloned.SetProperty(pr)

        # assign the same transformation to the copy
        cloned.SetOrigin(self.GetOrigin())
        cloned.SetScale(self.GetScale())
        cloned.SetOrientation(self.GetOrientation())
        cloned.SetPosition(self.GetPosition())

        cloned.mapper().SetScalarMode(self.mapper().GetScalarMode())
        cloned.name = self.name

        cloned.pipeline = utils.OperationNode(
            "clone", c="#edabab", shape="diamond", parents=[self],
        )
        return cloned

    def compute_quality(self, metric=7):
        """
        Calculate functions of quality for the elements of a triangular mesh.
        This method adds to the mesh a cell array named "Quality".

        Arguments:
            metric : (int)
                type of estimators:
                    - EDGE RATIO, 0
                    - ASPECT RATIO, 1
                    - RADIUS RATIO, 2
                    - ASPECT FROBENIUS, 3
                    - MIN_ANGLE, 4
                    - COLLAPSE RATIO, 5
                    - ASPECT GAMMA, 6
                    - VOLUME, 7
                    - ...

        See class [vtkMeshQuality](https://vtk.org/doc/nightly/html/classvtkMeshQuality.html)
        for an explanation of the meaning of each metric..
        """
        qf = vtk.vtkMeshQuality()
        qf.SetInputData(self._data)
        qf.SetTetQualityMeasure(metric)
        qf.SaveCellQualityOn()
        qf.Update()
        self._update(qf.GetOutput())
        return utils.vtk2numpy(qf.GetOutput().GetCellData().GetArray("Quality"))

    def compute_tets_volume(self):
        """Add to this mesh a cell data array containing the tetrahedron volume."""
        csf = vtk.vtkCellSizeFilter()
        csf.SetInputData(self._data)
        csf.SetComputeArea(False)
        csf.SetComputeVolume(True)
        csf.SetComputeLength(False)
        csf.SetComputeVertexCount(False)
        csf.SetVolumeArrayName("TetVolume")
        csf.Update()
        self._update(csf.GetOutput())
        return utils.vtk2numpy(csf.GetOutput().GetCellData().GetArray("TetVolume"))

    def check_validity(self, tol=0):
        """
        Return an array of possible problematic tets following this convention:
        ```python
        Valid               =  0
        WrongNumberOfPoints = 01
        IntersectingEdges   = 02
        IntersectingFaces   = 04
        NoncontiguousEdges  = 08
        Nonconvex           = 10
        OrientedIncorrectly = 20
        ```

        Arguments:
            tol : (float)
                This value is used as an epsilon for floating point
                equality checks throughout the cell checking process.
        """
        vald = vtk.vtkCellValidator()
        if tol:
            vald.SetTolerance(tol)
        vald.SetInputData(self._data)
        vald.Update()
        varr = vald.GetOutput().GetCellData().GetArray("ValidityState")
        return utils.vtk2numpy(varr)

    def threshold(self, name=None, above=None, below=None, on="cells"):
        """
        Threshold the tetrahedral mesh by a cell scalar value.
        Reduce to only tets which satisfy the threshold limits.

        - if `above = below` will only select tets with that specific value.
        - if `above > below` selection range is flipped.

        Set keyword "on" to either "cells" or "points".
        """
        th = vtk.vtkThreshold()
        th.SetInputData(self._data)

        if name is None:
            if self.celldata.keys():
                name = self.celldata.keys()[0]
                th.SetInputArrayToProcess(0, 0, 0, 1, name)
            elif self.pointdata.keys():
                name = self.pointdata.keys()[0]
                th.SetInputArrayToProcess(0, 0, 0, 0, name)
            if name is None:
                vedo.logger.warning("cannot find active array. Skip.")
                return self
        else:
            if on.startswith("c"):
                th.SetInputArrayToProcess(0, 0, 0, 1, name)
            else:
                th.SetInputArrayToProcess(0, 0, 0, 0, name)

        if above is not None and below is not None:
            if above > below:
                if vedo.vtk_version[0] >= 9:
                    th.SetInvert(True)
                    th.ThresholdBetween(below, above)
                else:
                    vedo.logger.error("in vtk<9, above cannot be larger than below. Skip")
                    return self
            else:
                th.ThresholdBetween(above, below)

        elif above is not None:
            th.ThresholdByUpper(above)

        elif below is not None:
            th.ThresholdByLower(below)

        th.Update()
        return self._update(th.GetOutput())

    def decimate(self, scalars_name, fraction=0.5, n=0):
        """
        Downsample the number of tets in a TetMesh to a specified fraction.
        Either `fraction` or `n` must be set.

        Arguments:
            fraction : (float)
                the desired final fraction of the total.
            n : (int)
                the desired number of final tets

        .. note:: setting `fraction=0.1` leaves 10% of the original nr of tets.
        """
        decimate = vtk.vtkUnstructuredGridQuadricDecimation()
        decimate.SetInputData(self._data)
        decimate.SetScalarsName(scalars_name)

        if n:  # n = desired number of points
            decimate.SetNumberOfTetsOutput(n)
        else:
            decimate.SetTargetReduction(1 - fraction)
        decimate.Update()
        self._update(decimate.GetOutput())
        self.pipeline = utils.OperationNode(
            "decimate", comment=f"array: {scalars_name}",
            c="#edabab", parents=[self],
        )
        return self

    def subdvide(self):
        """
        Increase the number of tets of a `TetMesh`.
        Subdivide one tetrahedron into twelve for every tetra.
        """
        sd = vtk.vtkSubdivideTetra()
        sd.SetInputData(self._data)
        sd.Update()
        self._update(sd.GetOutput())
        self.pipeline = utils.OperationNode(
            "subdvide", c="#edabab", parents=[self],
        )
        return self

    def isosurface(self, value=True):
        """
        Return a `vedo.Mesh` isosurface.

        Set `value` to a single value or list of values to compute the isosurface(s).
        """
        if not self._data.GetPointData().GetScalars():
            self.map_cells_to_points()
        scrange = self._data.GetPointData().GetScalars().GetRange()
        cf = vtk.vtkContourFilter()  # vtk.vtkContourGrid()
        cf.SetInputData(self._data)

        if utils.is_sequence(value):
            cf.SetNumberOfContours(len(value))
            for i, t in enumerate(value):
                cf.SetValue(i, t)
            cf.Update()
        else:
            if value is True:
                value = (2 * scrange[0] + scrange[1]) / 3.0
            cf.SetValue(0, value)
            cf.Update()

        clp = vtk.vtkCleanPolyData()
        clp.SetInputData(cf.GetOutput())
        clp.Update()
        msh = Mesh(clp.GetOutput(), c=None).phong()
        msh.mapper().SetLookupTable(utils.ctf2lut(self))
        msh.pipeline = utils.OperationNode("isosurface", c="#edabab", parents=[self])
        return msh

    def slice(self, origin=(0, 0, 0), normal=(1, 0, 0)):
        """
        Return a 2D slice of the mesh by a plane passing through origin and assigned normal.
        """
        strn = str(normal)
        if strn   ==  "x": normal = (1, 0, 0)
        elif strn ==  "y": normal = (0, 1, 0)
        elif strn ==  "z": normal = (0, 0, 1)
        elif strn == "-x": normal = (-1, 0, 0)
        elif strn == "-y": normal = (0, -1, 0)
        elif strn == "-z": normal = (0, 0, -1)
        plane = vtk.vtkPlane()
        plane.SetOrigin(origin)
        plane.SetNormal(normal)

        cc = vtk.vtkCutter()
        cc.SetInputData(self._data)
        cc.SetCutFunction(plane)
        cc.Update()
        msh = Mesh(cc.GetOutput()).flat().lighting("ambient")
        msh.mapper().SetLookupTable(utils.ctf2lut(self))
        msh.pipeline = utils.OperationNode("slice", c="#edabab", parents=[self])
        return msh



import numpy as np
import napari
import vedo

# Load the surface, triangulate just in case, and compute vertex normals
surf = vedo.Mesh(vedo.dataurl+"beethoven.ply").triangulate().compute_normals()
surf.rotate_x(180).rotate_y(60)

vertices = surf.points()
faces    = np.array(surf.faces())
normals  = surf.normals()
# generate vertex values by projecting normals on a "lighting vector"
values   = np.dot(normals, [-1, 1, 1])

print(vertices.shape, faces.shape, values.shape)
# (2521, 3) (5030, 3) (2521,)

with napari.gui_qt():
    # create an empty viewer
    viewer = napari.Viewer()

    # add the surface
    viewer.add_surface((vertices, faces, values), opacity=0.8)
    viewer.add_points(vertices, size=0.05, face_color='pink')

    # turn on 3D rendering
    viewer.dims.ndisplay = 3



"""Erode or dilate a Volume
by replacing a voxel with the max/min
over an ellipsoidal neighborhood"""
from vedo import *

embryo = Volume(dataurl+'embryo.tif')
embryo.print_histogram(logscale=1)

eroded = embryo.clone().erode(neighbours=(2,2,2))
dilatd = eroded.clone().dilate(neighbours=(2,2,2))

show([(embryo, __doc__), eroded, dilatd], N=3, viewup='z', zoom=1.4).close()




"""Forward kinematics: hover the mouse to drag the chain"""
from vedo import Plotter, versor, Plane, Line

n = 15  # number of points
l =  3  # length of one segment

def func(evt):
    if not evt.actor:
        return
    coords = line.points()
    coords[0] = evt.picked3d
    for i in range(1, n):
        v = versor(coords[i] - coords[i-1])
        coords[i] = coords[i-1] + v * l
    line.points(coords)  # update positions
    nodes.points(coords)
    plt.render()

surf = Plane(s=[60, 60])
line = Line([l*n/2, 0], [-l*n/2, 0], res=n, lw=12)
nodes= line.clone().c('red3').point_size(15)

plt = Plotter()
plt.add_callback("on mouse move please call", func)
plt.show(surf, line, nodes, __doc__, zoom=1.3)
plt.close()



"""Categories and repeats"""
# manually create a plot by adding Rectangles to a Figure object
from vedo import np, settings, Rectangle, Text3D, Line, DashedLine
from vedo.pyplot import Figure

settings.default_font = "Theemim"

#################################################################### First plot
groupA = np.random.randn(3)*10+50
groupB = np.random.randn(3)*10+60
groupC = np.random.randn(3)*10+70

fig = Figure(
    [-5,55], [-10,100],   # x and y ranges
    xtitle='', ytitle='', # this disables x and y axes
)

#################
x0 = 0
for i in range(3):
    x1 = x0 + 4
    val= groupA[i]
    fig += Rectangle([x0,0], [x1, val], c=f'red{3+i*2}')
    x0 = x1
fig += Text3D("Group A", justify='center', c='k').pos(6,-7).scale(4)
fig += Line([-1,0], [13, 0], lw=2)

#################
x0 = 20
for i in range(3):
    x1 = x0 + 4
    val= groupB[i]
    fig += Rectangle([x0,0], [x1, val], c=f'purple{3+i*2}')
    x0 = x1
fig += Text3D("Group B", justify='center', c='k').pos(26,-7).scale(4)
fig += Line([19,0], [33, 0], lw=2)

#################
x0 = 40
for i in range(3):
    x1 = x0 + 4
    val= groupC[i]
    fig += Rectangle([x0,0], [x1, val], c=f'orange{3+i*2}')
    x0 = x1
fig += Text3D("Group C", justify='center', c='k').pos(46,-7).scale(4)
fig += Line([39,0], [53, 0], lw=2)

#################
fig += DashedLine([-2,50], [55,50], c='k3', lw=1)
fig += Text3D("50%").pos(-7,49).scale(3).c('k')
fig.show(size=(1000,700), zoom='tight', title=__doc__).clear()


#################################################################### Second plot
fig = Figure(
    [0, 100], [-20, 80],
    aspect=3/4,             # can change the aspect ratio
    xtitle='', ytitle='', # this disables x and y axes
)
for i in range(5):
    val = np.random.randn()*10+50
    y0, y1 = 2*i, 2*i+1
    fig += Rectangle([0,y0], [100,y1], radius=0.5, c='k6')
    fig += Rectangle([0,y0], [val,y1], radius=0.5, c='r4').z(1)
fig += Text3D("Group A", justify='center', c='k').pos(50,-5).scale(2.5)


for i in range(5):
    val = np.random.randn()*10+60
    y0, y1 = 2*i + 20, 2*i+1 + 20
    fig += Rectangle([0,y0], [100,y1], radius=0.5, c='k6')
    fig += Rectangle([0,y0], [val,y1], radius=0.5, c='p5').z(1)
fig += Text3D("Group B", justify='center', c='k').pos(50,15).scale(2.5)

for i in range(5):
    val = np.random.randn()*10+70
    y0, y1 = 2*i + 40, 2*i+1 + 40
    fig += Rectangle([0,y0], [100,y1], radius=0.5, c='k6')
    fig += Rectangle([0,y0], [val,y1], radius=0.5, c='o5').z(1)
fig += Text3D("Group C", justify='center', c='k').pos(50,35).scale(2.5)

fig.show(size=(1000,700), zoom='tight').close()



"""Draw the shadow and trailing line of a moving object."""
from vedo import *

world = Box(size=(30,15,8)).wireframe()
airplane = Mesh(dataurl+"cessna.vtk").c("green")
airplane.pos(-15, 2.0, 0.15)
airplane.add_trail(n=100).add_shadow('z', -4)

plt = Plotter(interactive=False)
plt.show(world, airplane, __doc__, viewup="z")

for t in np.arange(0, 3.2, 0.04):
    pos = (9*t-15, 2-t, sin(3-t))  # make up some movement
    airplane.pos(pos).rotate_x(t)
    airplane.update_trail()
    airplane.update_shadows()
    plt.render()

plt.interactive().close()

